
/*
  
   GNS EDR Structure definitions -- Concentrator Housekeeping Data 
  
   BVS/NAIF, June 7, 2000
  
      Changed sclk_subseconds type from uint8 to uint16 to preserve 
      full resolution of the fractional part of SCLK. Updated 
      description to reflect this change. Added APID_status, SW_status
      and CMD_status bit breakdown description.
        
   BVS/NAIF, March 2, 2000
  
      Reformatted comments; included gnd_edr_const.h.
  
   BVS/NAIF, February 17, 2000
  
*/

#include "hdfi.h"
#include "gns_edr_const.h"

/*
   
   Concentrator Housekeeping data structure -- APID # 25
   
      sclk_seconds      -- SCLK seconds 

      sclk_subseconds   -- fractions (1/65536th) of SCLK second 

      APID_status       -- APID status word (? bit order):
      
                              bit 0 = Checksum_Error_Occ_Indication
                              bit 1 = Sample_Mode
                              bits 2-3 = COLLECTION_TYPE

      elec_req          -- from last config command sent

      accel_req         -- from last config command sent

      rej_req           -- from last config command sent

      elec_level        -- ADC value from electrode voltage readback, 
                           (0-4095d)

      accel_level       -- ADC value from acceleration voltage readback,
                           (0-4095d)

      rej_level         -- ADC value from rejection voltage readback, 
                           (0-4095d)

      elec_set          -- ADC value from electrode voltage DAC 
                           readback, (0-4095d)

      accel_set         -- ADC value from acceleration voltage DAC 
                           readback, (0-4095d)

      rej_set           -- ADC value from rejection voltage DAC 
                           readback, (0-4095d)

      MON_5V            -- ADC value from 5 volt monitor, (0-4095d)                        

      MON_12V           -- ADC value from 12 volt monitor, (0-4095d)                        

      elec_fault_count  -- count of electrode voltage errors since 
                           last report, (0-65535d)

      accel_fault_count -- count of acceleration voltage errors since 
                           last report, (0-65535d)

      rej_fault_count   -- count of rejection voltage errors since 
                           last report, (0-65535d)

      elec_maximum      -- maximum electrode voltage readback value 
                           since last report, (0-4095d)

      elec_minimum      -- minimum electrode voltage readback value 
                           since last report, (0-4095d)

      accel_maximum     -- maximum acceleration voltage readback value 
                           since last report, (0-4095d)

      accel_minimum     -- minimum acceleration voltage readback value 
                           since last report, (0-4095d)

      rej_maximum       -- maximum rejection voltage readback value 
                           since last report, (0-4095d)

      rej_minimum       -- minimum rejection voltage readback value 
                           since last report, (0-4095d)

      SW_status         -- Concentrator SW status word (? bit order):              

                              bit 0 = HV_ENABLE
                              bit 1 = HV_ENABLE_LIMIT
                              bit 2 = RAM_TEST
                              bit 3 = ROM_TEST
                              bit 4 = INTERUPT_TEST
                              bit 5 = TIMING_TEST
                              bit 6 = DAC1_TEST
                              bit 7 = DAC2_TEST
                              bit 8 = DAC3_TEST
                              bit 9 = PROCESSOR_RESET

      CMD_status        -- Concentrator CMD status word (? bit order):
                    
                              bits 0-7 = Command_Status_Counter
                              bit 8 = CMD_Status_Bad_Command_Flag
                              bit 9 = CMD_Status_Bad_Parameter_Flag

*/

struct Concentrator_Housekeeping {

   uint32   sclk_seconds;
   uint16   sclk_subseconds;
   uint16   APID_status;
   uint16   elec_req;
   uint16   accel_req;
   uint16   rej_req;
   uint16   elec_level;
   uint16   accel_level;
   uint16   rej_level;
   uint16   elec_set;
   uint16   accel_set;
   uint16   rej_set;
   uint16   MON_5V;
   uint16   MON_12V;
   uint16   elec_fault_count;
   uint16   accel_fault_count;
   uint16   rej_fault_count;
   uint16   elec_maximum;
   uint16   elec_minimum;
   uint16   accel_maximum;
   uint16   accel_minimum;
   uint16   rej_maximum;
   uint16   rej_minimum;
   uint16   SW_status;
   uint16   CMD_status;
                                       
};


