
/*
   Prototype GNS EDR production program
   
   Version 4.1.0 -- BVS/NAIF -- April 12, 2001
   
      Modified MEC/WIND HK packet (APID 24) decom to accomodate 
      new field added to the packet (LMA class I/II change ref 
      No. XH0304 from March 2001.)
   
   Version 4.0.0 -- BVS/NAIF -- November 21, 2000
   
      Added processing for GEM and GIM raw count packets (APIDs 41 
      and 42).
   
   Version 3.0.0 -- BVS/NAIF -- August 28, 2000
   
      Added processing for SPIN TRACK quaternion packet (APID 27).
      Made specification of the instrument name on the command line
      required.
   
   Version 2.0.0 -- BVS/NAIF -- August 17, 2000
   
      Added processing for MEC/WIND F1D APID (APID 26).
      Changed packing for the last field in MEC_HK to package float32.
   
   Version 1.1.0 -- BVS/NAIF -- July 7, 2000
   
      Changed SCLK subsecond packing to use two bytes in all 
      HDF structure pack routines. Added processing for multiple 
      data records in a single packet for all APID except GEM_Normal,
      GIM_Normal and GIM_Diag. Replaced printing of error and warning
      messages to stderr with stdout. Consistently assigned exit 
      codes.

   Version 1.0.3 -- BVS/NAIF -- June 16, 2000
   
      Added processing for GIM/Diag. Added packet size and 
      segment consistency checks.

   Version 1.0.2 -- BVS/NAIF -- May 7, 2000
   
      Updated MEC/WIND packet code.

   Version 1.0.1 -- BVS/NAIF -- April 11, 2000
   
      Very draft version working of test TLM data.

   Version 1.0.0 -- BVS/NAIF -- March 8, 2000
   
      Initial version producing EDRs with manufactured data for back
      end testing.

*/
#define PROGNAME  "gns_edr"
#define PROGVER   "Version 4.1.0 -- BVS/NAIF -- April 12, 2001"

#include <stdio.h>
#include <math.h>
#include <unistd.h>
#include <sys/param.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/file.h>
#include <fcntl.h>

/*
   HDF stuff
*/
#include "hdfi.h"
#include "df.h"
#include "mfhdf.h"

/*
   GNS EDR constants and structure stuff
*/
#include "gns_edr_const.h"
#include "con_hk.h"
#include "gem_hk.h"
#include "gem_mancal.h"
#include "gem_normal.h"
#include "gem_raw.h"
#include "gim_diag.h"
#include "gim_hk.h"
#include "gim_mancal.h"
#include "gim_normal.h"
#include "gim_raw.h"
#include "mec_hk.h"
#include "mec_f1d.h"

/*
   Local Constants
*/

/*
   Instruments IDs.
*/
#define GEM 1
#define GIM 2
#define CON 3
#define MEC 4
#define ACS 5

/*
   APID IDs
*/
#define GEM_HK_APID             32
#define GEM_MANCAL_APID         33
#define GEM_NORMAL_APID         31
#define GEM_RAW_APID            41
#define GIM_HK_APID             29
#define GIM_MANCAL_APID         30
#define GIM_NORMAL_APID         28
#define GIM_DIAG_APID           34
#define GIM_RAW_APID            42
#define CON_HK_APID             25
#define MEC_HK_APID             24
#define MEC_F1D_APID            26
#define STRK_QUAT_APID          27

/*  
   
   APID sizes (in bytes :-)
   
*/ 
#define GEM_HK_SIZE             46        
#define GEM_NORMAL_SIZE       3454  
#define GEM_MANCAL_SIZE         50     
#define GEM_RAW_SIZE          2360     
#define GIM_HK_SIZE             66  
#define GIM_NORMAL_SIZE       3336  
#define GIM_MANCAL_SIZE         94      
#define GIM_DIAG_SIZE        12934
#define GIM_RAW_SIZE          8328
#define CON_HK_SIZE             52      
#define MEC_HK_SIZE            144
#define MEC_F1D_SIZE           168
#define STRK_QUAT_SIZE          40

/*
   Packet "chunk" types.
*/
#define FIRST                    1
#define CONTINUATION             0
#define LAST                     2
#define SOLE                     3

/*
   SFDU stuff
*/
#define LABEL_SIZE              20
#define CHDO_SIZE                4
#define MAX_SPL_REC          32768
#define HEADER                  20
#define AGGREGATE                4
#define DATA_CHDO                4

/* 
   CHDO 82 Information (from Howard's dma.h)
*/
#define CHDO82_SCID                 6
#define CHDO82_SCID_MASK               0xFF
#define CHDO82_ERT                  10
#define CHDO82_ERT_MASK                0xFFFFFFFFFFFF
#define CHDO82_RCT                  64 
#define CHDO82_RCT_MASK                0xFFFFFFFFFFFF
#define CHDO82_ANOMALY_FLAGS        70
#define CHDO82_ANOMALY_FLAGS_MASK      0xFFFF

/* 
   CHDO 83 Information (from Howard's dma.h) 
*/
#define CHDO83_PKT_FILL_FLAG        4
#define CHDO83_PKT_FILL_FLAG_MASK      0x80
#define CHDO83_PACKET_APID          4
#define CHDO83_PACKET_APID_MASK        0x07FF
#define CHDO83_SRC_PKT_SEQ_COUNT    6
#define CHDO83_SRC_PKT_SEQ_COUNT_MASK  0xFFFF
#define CHDO83_NON_FILL_LENGTH      8
#define CHDO83_NON_FILL_LENGTH_MASK    0xFFFF
#define CHDO83_SCLK                 14
#define CHDO83_SCLK_MASK               0xFFFFFFFFFFFF
#define CHDO83_SCET                 20
#define CHDO83_SCET_MASK               0xFFFFFFFFFFFF

/*
   Numeric re-packaging stuff.
*/
#define U16 unsigned short
#define U32 unsigned int 
#define C16_TO_U16(cp)    (*((U16 *) (cp)))
#define C32_TO_U32(cp)    (*((U32 *) (cp)))
#define C32_TO_F32(cp, f) (*((float *) memcpy((char *) &(f), (cp), 4)))
#define C64_TO_F64(cp, d) (*((double *) memcpy((char *) &(d), (cp), 8)))

/*
   2/4 byte packaging for bigendian.
*/
#define U8_TO_U32(A,B,C,D)  ((A<<24)+(B<<16)+(C<<8)+(D))
#define U8_TO_U16(A,B)      ((A<<8)+(B))

/*
   Command line keys and reserved values.
*/
#define CLKINST "-inst"
#define CLKGEM  "gem"
#define CLKGIM  "gim"
#define CLKCON  "con"
#define CLKMEC  "mec"
#define CLKACS  "acs"

#define CLKAPID "-apid"
#define CLKNORM "normal"
#define CLKHK   "hk"
#define CLKRAW  "raw"
#define CLKMAN  "mancal"
#define CLKDIAG "diag"
#define CLKF1D  "f1d"
#define CLKQUAT "quats"

#define CLKTLMF "-tlm"
#define CLKEDRF "-edr"




int main( int argc, char * argv[] )
{

   /*
   Current APID ID, packet segment type and instrument ID.
   */
   int apid_ID;
   int instrument;
   int segment_type;
   
   /*
   Requested instrument and APID IDs for instrument/APID specific 
   runs.
   */
   int requested_apid         = 0;
   int requested_instrument   = 0;
   
   /*
   Logical flag that control text file initialization.
   */
   int TXT_opened             = 0;
   
   /*
   Logical flags that control HDF initialization.
   */
   int HDF_opened             = 0;
   int GEM_HK_started         = 0;
   int GEM_MANCAL_started     = 0;
   int GEM_NORMAL_started     = 0;
   int GEM_RAW_started        = 0;
   int GIM_HK_started         = 0;
   int GIM_MANCAL_started     = 0;
   int GIM_NORMAL_started     = 0;
   int GIM_DIAG_started       = 0;
   int GIM_RAW_started        = 0;
   int CON_HK_started         = 0;
   int MEC_HK_started         = 0;
   int MEC_F1D_started        = 0;
   
   /*
   Flags that control long packet assembly.
   */
   int GEM_NORMAL_assembled   = 0;
   int GEM_RAW_assembled      = 0;
   int GIM_NORMAL_assembled   = 0;
   int GIM_DIAG_assembled     = 0;
   int GIM_RAW_assembled      = 0;
   
   /*
   HDF file name and HDF file and data type IDs (something a-la SPICE 
   file ``HANDLE''s :-).
   */
   char   tlm_file[MAXPATHLEN];
   char   hdf_file[MAXPATHLEN];
   char   tmp_file[MAXPATHLEN];
   int32  hdf_fp_out;
   int32  hdf_sd_out;
   int32  hdf_an_out;
   FILE  *txt_fp_out;
   
   /*
   Data segment and assembled packet buffers and sizes.
   */
   int sfdu_size;
   int sfdu_sec;
   int sfdu_subsec;
   int sfdu_seqcnt;
   
   uint8 sfdu[MAX_SPL_REC];
   
   int gem_norm_size;
   int gem_raw_size;
   int gim_norm_size;
   int gim_diag_size;
   int gim_raw_size;
   
   uint8 gem_norm_sfdu[MAX_SPL_REC];
   uint8 gem_raw_sfdu[MAX_SPL_REC];
   uint8 gim_norm_sfdu[MAX_SPL_REC];
   uint8 gim_diag_sfdu[MAX_SPL_REC];
   uint8 gim_raw_sfdu[MAX_SPL_REC];
   
   /*
   Packet times for segmented packets.
   */
   int gem_norm_sec    = 0;
   int gem_norm_subsec = 0;
   int gem_raw_sec     = 0;
   int gem_raw_subsec  = 0;
   int gim_norm_sec    = 0;
   int gim_norm_subsec = 0;
   int gim_diag_sec    = 0;
   int gim_diag_subsec = 0;
   int gim_diag_seqcnt = 0;
   int gim_raw_sec     = 0;
   int gim_raw_subsec  = 0;
   int gim_raw_seqcnt  = 0;
   
   /*
   Misc. variables.
   */
   int    i;
   int    clkey;
   int    sclk_sec;
   int    sclk_sub;
   int    quatoff;
   double quat[4];
   double quattim;
   double quatsec;
   double quatsub;

   
   /*
   Parse command line.
   */
   parse_command_line( argc, argv, 
                       &requested_instrument,
                       &requested_apid,
                       tlm_file,
                       hdf_file  );
                       
   printf( "\n" );
   printf( "Command line:\n\n" );
   printf( "   inst = %i\n", requested_instrument );
   printf( "   apid = %i\n", requested_apid );
   printf( "   tlmf = %s\n", tlm_file );
   printf( "   edrf = %s\n", hdf_file );
   printf( "\n" );
   
   /*
   We will follow Ed Santiago's suggestion and write to a temporary 
   HDF file during the run. If the file will be created OK, then we
   will move it to the actual file at the end of the program.
   */
   sprintf( tmp_file, "%s.tmp", hdf_file );
   
   /*
   This is the main loop that pulls SFDU wrapped packets, one by one, 
   from input raw packet file.
   */
   
   while ( got_good_apid(  tlm_file,
                           &apid_ID,
                           &segment_type,
                           &sfdu_size,
                           &sfdu_sec,
                           &sfdu_subsec,
                           &sfdu_seqcnt,
                           sfdu   ) 
         )
   {     

      /*
      Check whether we are interested in a specific instrument 
      during this run.
      */
      if ( requested_instrument == 0 )
      {
         /*
         No, in this run we process data for all instruments. Assign
         instrument ID based on the APID.
         */
         switch ( apid_ID )
         {
            case GEM_HK_APID: 
            case GEM_MANCAL_APID: 
            case GEM_NORMAL_APID:
            case GEM_RAW_APID:
               instrument = GEM;
               break;
               
            case GIM_HK_APID:
            case GIM_MANCAL_APID:
            case GIM_NORMAL_APID:
            case GIM_DIAG_APID:
            case GIM_RAW_APID:
               instrument = GIM;
               break;
               
            case CON_HK_APID:
               instrument = CON;
               break;
               
            case MEC_HK_APID:
            case MEC_F1D_APID:
               instrument = MEC;
               break;
               
            case STRK_QUAT_APID:
               instrument = ACS;
               break;
         }
      }
      else 
      {
         instrument = requested_instrument;
      }
   
      /*
      Check whether we also were requested to work only with specific 
      APID and our current APID matches it.
      */
      
      if ( ( requested_apid != 0 ) && ( apid_ID != requested_apid ) )
      {
         /*
         It does not match it. We reset APID ID to make sure that the 
         rest of the code simply skips this packet.
         */
         apid_ID = 0;
      }
      
      /*
      Check whether we have valid instrument/APID pair but didn't open
      the output HDF file (for GEM/GIM/CON/MEC) data or text file for 
      SPIN TRACK quaternion data yet. If so, open the file (note that  
      this "if" is not "perfect" because even if we have correct
      combination of instrument and APID, for normal and diag packets 
      we may have only a chunk of what we need.)
      */
      if ( (( instrument == GEM && apid_ID == GEM_HK_APID )     ||
            ( instrument == GEM && apid_ID == GEM_MANCAL_APID ) ||
            ( instrument == GEM && apid_ID == GEM_NORMAL_APID ) ||
            ( instrument == GEM && apid_ID == GEM_RAW_APID )    ||
            ( instrument == GIM && apid_ID == GIM_HK_APID )     ||
            ( instrument == GIM && apid_ID == GIM_MANCAL_APID ) ||
            ( instrument == GIM && apid_ID == GIM_NORMAL_APID ) ||
            ( instrument == GIM && apid_ID == GIM_DIAG_APID )   ||
            ( instrument == GIM && apid_ID == GIM_RAW_APID )    ||
            ( instrument == CON && apid_ID == CON_HK_APID )     ||
            ( instrument == MEC && apid_ID == MEC_HK_APID )     ||
            ( instrument == MEC && apid_ID == MEC_F1D_APID )      ) &&
              HDF_opened == 0 )
      {
         /*
         It looks like we have got first packet that needs to be written
         to EDR. Time to open output EDR file. The open_EDR_file does
         some magid HDF related stuff and return three HDF ``ID'' for 
         Vdata, SD and Annotations.
         */
         open_HDF_file ( tmp_file, 
                         &hdf_fp_out, 
                         &hdf_sd_out, 
                         &hdf_an_out );
                         
         HDF_opened = 1;
         
         printf( "opened %s\n", tmp_file );
         
      }
      else if ( (( instrument == ACS && apid_ID == STRK_QUAT_APID )) &&
                   TXT_opened == 0 )
      {
         /*
         Open text file for quaternion output and set the flag 
         indicating that it was opened.
         */
         txt_fp_out = fopen( tmp_file, "w" );
         if ( txt_fp_out == NULL ) 
         {
            printf(  "ERROR: Cannot open output file %s.\n", tmp_file );
            exit(-1);
         }
         
         TXT_opened = 1;
         
         printf( "opened %s\n", tmp_file );
         
      }
      
      /*
      At this point we know that current SFDU is for the 
      instrument/APID of our interest. We pulls it though this case 
      which contains specific actions to be done for each packet.
      */
      switch ( instrument ) 
      {
      
         /*
         Process GEM data.
         */
         case GEM:
   
            switch ( apid_ID )
            {
         
               case GEM_HK_APID:
               
                  /*
                  Check packet size -- must be greater than zero and
                  multiple of elementary data chunks. If OK, process 
                  the packet.
                  */
                  if ( sfdu_size > 0                && 
                       sfdu_size % GEM_HK_SIZE == 0     ) 
                  {
                     /*
                     Initialize GEM h/k HDF structures if needed.
                     */
                     if ( ! GEM_HK_started )
                     {
                        if ( init_wr_gem_hk ( hdf_fp_out, 
                                              hdf_sd_out, 
                                              hdf_an_out, 
                                              "tlm_edr_data" ) == FAIL )
                        {
                           printf(  
                                 "ERROR: init_wr_gem_hk failed\n" );
                           exit(-1);
                        } 
                        else 
                        {
                           GEM_HK_started = 1;
   
                           printf( "   initialized GEM h/k\n" );
                        }
                     }
                     
                     /*
                     GEM h/k is a sol packet but it can contain 
                     multiple data chunks -- call function that 
                     packages each into a structure and writes it  
                     to the file.
                     */
                     for ( i = 0; i < sfdu_size / GEM_HK_SIZE; i++ ) 
                     {
                        put_GEM_hk( &sfdu[0 + i*GEM_HK_SIZE] );
                     }
                     
                     /*
                     Current GEM h/k packet has been processed.
                     */
                  }
                  else
                  {
                     printf(  
                        "WARNING: dropped GEM HK @%10d:%03d -- bad size %d\n",
                        sfdu_sec, sfdu_subsec, sfdu_size );
                  }
                  
                  break;
            
               case GEM_MANCAL_APID:
               
                  /*
                  Check packet size -- must be greater than zero and
                  multiple of elementary data chunks. If OK, process 
                  the packet.
                  */
                  if ( sfdu_size > 0                && 
                       sfdu_size % GEM_MANCAL_SIZE == 0     ) 
                  {
                     /*
                     Initialize GEM man/cal HDF structures if needed.
                     */
                     if ( ! GEM_MANCAL_started )
                     {
                        if ( init_wr_gem_mancal ( hdf_fp_out, 
                                              hdf_sd_out, 
                                              hdf_an_out, 
                                              "tlm_edr_data" ) == FAIL )
                        {
                           printf( 
                              "ERROR: init_wr_gem_mancal failed\n");
                           exit(-1);
                        } 
                        else 
                        {
                           GEM_MANCAL_started = 1;
                           
                           printf( "   initialized GEM man/cal\n" );
                        }
                        
                     }
                     
                     /*
                     GEM man/cal is a sol packet but it can contain 
                     multiple data chunks -- call function that 
                     packages each into a structure and writes it  
                     to the file.
                     */
                     for ( i = 0; i < sfdu_size / GEM_MANCAL_SIZE; i++ ) 
                     {
                        put_GEM_mancal( &sfdu[0 + i * GEM_MANCAL_SIZE]);
                     }
                     
                     /*
                     Current GEM man/cal packet has been processed.
                     */
                  }
                  else
                  {
                     printf(  
                        "WARNING: dropped GEM ManCal @%10d:%03d -- bad size %d\n",
                        sfdu_sec, sfdu_subsec, sfdu_size );
                  }
                  
                  break;
            
               case GEM_NORMAL_APID:
               
                  /*
                  GEM normal is a split packet. Since the packet was
                  split into two parts, handle the data block
                  depending on whether it's first or last chuck.
                  */
                  switch ( segment_type )
                  {
                     case FIRST:
                        
                        /*
                        This is the first chunk. We copy it to the 
                        buffer that holds assembled packet. We also 
                        save size of the packet and packet times (we 
                        will use times in checks of continution and 
                        last chunks.)
                        */
                        memmove( gem_norm_sfdu, sfdu, sfdu_size );
                        gem_norm_size   = sfdu_size;
                        gem_norm_sec    = sfdu_sec;
                        gem_norm_subsec = sfdu_subsec;
                        break;
                                                
                     case LAST:
                     
                        /*
                        This is the second chunk. If it has the same 
                        time, we copy it to the buffer that holds 
                        assembled packet. Otherwise we report two 
                        "orphan" packets.
                        */
                        if ( gem_norm_sec    == sfdu_sec    && 
                             gem_norm_subsec == sfdu_subsec    )
                        {
                           memmove( &gem_norm_sfdu[gem_norm_size], 
                                    sfdu, 
                                    sfdu_size );
                           gem_norm_size += sfdu_size;
                           GEM_NORMAL_assembled = 1;
                        
                           printf( "accumulated GEM normal size %d\n", 
                                    gem_norm_size);
                        } 
                        else 
                        {
                           printf(  
                              "WARNING: dropped orphan GEM_Norm FIRST @%10d:%03d\n",
                              gem_norm_sec, gem_norm_subsec );
                           printf(  
                              "WARNING: dropped orphan GEM_Norm LAST  @%10d:%03d\n",
                              sfdu_sec, sfdu_subsec );
                        }

                        break;
                        
                  };
                  
                  /*
                  Check whether we have collected all pieces of one 
                  packet.
                  */
                  if ( GEM_NORMAL_assembled )
                  {
                  
                     /*
                     Check packet size. If OK, process the packet.
                     */
                     if ( gem_norm_size == GEM_NORMAL_SIZE ) 
                     {
                        /*
                        Initialize GEM normal HDF structures if needed.
                        */
                        if ( ! GEM_NORMAL_started )
                        {
                           if ( init_wr_gem_normal ( hdf_fp_out, 
                                             hdf_sd_out, 
                                             hdf_an_out, 
                                             "tlm_edr_data" ) == FAIL )
                           {
                              printf(  
                                "ERROR: init_wr_gem_normal failed\n" );
                              exit(-1);
                           } 
                           else 
                           {
                              GEM_NORMAL_started = 1;
                           
                              printf( "   initialized GEM Normal\n" );
                           }
                        }
                        
                        /*
                        Package and write current data block.
                        */
                        put_GEM_normal ( gem_norm_sfdu );
                     }
                     else
                     {
                        printf(  
                           "WARNING: dropped GEM Normal @%10d:%03d -- bad size %d\n",
                           sfdu_sec, sfdu_subsec, gem_norm_size );
                     }
                                          
                     /*
                     Current GEM normal packet has been processed (or 
                     dropped. :-)
                     */
                     GEM_NORMAL_assembled = 0;
                  }
                  
                  break;

               case GEM_RAW_APID:
               
                  /*
                  GEM raw is a split packet. Since the packet was
                  split into two parts, handle the data block
                  depending on whether it's first or last chuck.
                  */
                  switch ( segment_type )
                  {
                     case FIRST:
                        
                        /*
                        This is the first chunk. We copy it to the 
                        buffer that holds assembled packet. We also 
                        save size of the packet and packet times (we 
                        will use times in checks of continution and 
                        last chunks.)
                        */
                        memmove( gem_raw_sfdu, sfdu, sfdu_size );
                        gem_raw_size   = sfdu_size;
                        gem_raw_sec    = sfdu_sec;
                        gem_raw_subsec = sfdu_subsec;
                        break;
                                                
                     case LAST:
                     
                        /*
                        This is the second chunk. If it has the same 
                        time, we copy it to the buffer that holds 
                        assembled packet. Otherwise we report two 
                        "orphan" packets.
                        */
                        if ( gem_raw_sec    == sfdu_sec    && 
                             gem_raw_subsec == sfdu_subsec    )
                        {
                           memmove( &gem_raw_sfdu[gem_raw_size], 
                                    sfdu, 
                                    sfdu_size );
                           gem_raw_size += sfdu_size;
                           GEM_RAW_assembled = 1;
                        
                           printf( "accumulated GEM raw size %d\n", 
                                    gem_raw_size);
                        } 
                        else 
                        {
                           printf(  
                              "WARNING: dropped orphan GEM_Raw FIRST @%10d:%03d\n",
                              gem_raw_sec, gem_raw_subsec );
                           printf(  
                              "WARNING: dropped orphan GEM_Raw LAST  @%10d:%03d\n",
                              sfdu_sec, sfdu_subsec );
                        }

                        break;
                        
                  };
                  
                  /*
                  Check whether we have collected all pieces of one 
                  packet.
                  */
                  if ( GEM_RAW_assembled )
                  {
                  
                     /*
                     Check packet size. If OK, process the packet.
                     */
                     if ( gem_raw_size == GEM_RAW_SIZE ) 
                     {
                        /*
                        Initialize GEM raw HDF structures if needed.
                        */
                        if ( ! GEM_RAW_started )
                        {
                           if ( init_wr_gem_raw ( hdf_fp_out, 
                                             hdf_sd_out, 
                                             hdf_an_out, 
                                             "tlm_edr_data" ) == FAIL )
                           {
                              printf(  
                                "ERROR: init_wr_gem_raw failed\n" );
                              exit(-1);
                           } 
                           else 
                           {
                              GEM_RAW_started = 1;
                           
                              printf( "   initialized GEM Raw\n" );
                           }
                        }
                        
                        /*
                        Package and write current data block.
                        */
                        put_GEM_raw ( gem_raw_sfdu );
                     }
                     else
                     {
                        printf(  
                           "WARNING: dropped GEM Raw @%10d:%03d -- bad size %d\n",
                           sfdu_sec, sfdu_subsec, gem_raw_size );
                     }
                                          
                     /*
                     Current GEM raw packet has been processed (or 
                     dropped. :-)
                     */
                     GEM_RAW_assembled = 0;
                  }
                  
                  break;
                  
            };
            break;
            
         /*
         Process GIM data.
         */
         case GIM:
   
            switch ( apid_ID )
            {
         
               case GIM_HK_APID:
               
                  /*
                  Check packet size -- must be greater than zero and
                  multiple of elementary data chunks. If OK, process 
                  the packet.
                  */
                  if ( sfdu_size > 0                && 
                       sfdu_size % GIM_HK_SIZE == 0     ) 
                  {
                     /*
                     Initialize GIM h/k HDF structures if needed.
                     */
                     if ( ! GIM_HK_started )
                     {
                        if ( init_wr_gim_hk ( hdf_fp_out, 
                                              hdf_sd_out, 
                                              hdf_an_out, 
                                              "tlm_edr_data" ) == FAIL )
                        {
                           printf(  
                              "ERROR: init_wr_gim_hk failed\n" );
                           exit(-1);
                        } 
                        else 
                        {
                           GIM_HK_started = 1;
                        
                           printf( "   initialized GIM h/k\n" );
                        }
                     }
                     
                     /*
                     GIM h/k is a sol packet but it can contain 
                     multiple data chunks -- call function that 
                     packages each into a structure and writes it  
                     to the file.
                     */
                     for ( i = 0; i < sfdu_size / GIM_HK_SIZE; i++ ) 
                     {
                        put_GIM_hk( &sfdu[0 + i * GIM_HK_SIZE]);
                     }

                     /*
                     Current GIM h/k packet has been processed.
                     */
                  }
                  else
                  {
                     printf( 
                        "WARNING: dropped GIM HK @%10d:%03d -- bad size %d\n",
                        sfdu_sec, sfdu_subsec, sfdu_size );
                  }
                  
                  break;
               
               case GIM_MANCAL_APID:
   
                  /*
                  Check packet size -- must be greater than zero and
                  multiple of elementary data chunks. If OK, process 
                  the packet.
                  */
                  if ( sfdu_size > 0                && 
                       sfdu_size % GIM_MANCAL_SIZE == 0     ) 
                  { 
                     /*
                     Initialize GIM man/cal HDF structures if needed.
                     */
                     if ( ! GIM_MANCAL_started )
                     {
                        if ( init_wr_gim_mancal ( hdf_fp_out, 
                                             hdf_sd_out, 
                                             hdf_an_out, 
                                             "tlm_edr_data" ) == FAIL )
                        {
                           printf(  
                              "ERROR: init_wr_gim_mancal failed\n");
                           exit(-1);
                        } 
                        else 
                        {
                           GIM_MANCAL_started = 1;
                        
                           printf( "   initialized GIM Man/Cal\n" );
                        }
                     }
                     
                     /*
                     GIM man/cal is a sol packet but it can contain 
                     multiple data chunks -- call function that 
                     packages each into a structure and writes it  
                     to the file.
                     */
                     for ( i = 0; i < sfdu_size / GIM_MANCAL_SIZE; i++ ) 
                     {
                        put_GIM_mancal( &sfdu[0 + i * GIM_MANCAL_SIZE]);
                     }
                     
                     /*
                     Current GIM man/cal packet has been processed.
                     */
                  }
                  else
                  {
                     printf( 
                        "WARNING: dropped GIM ManCal @%10d:%03d -- bad size %d\n",
                        sfdu_sec, sfdu_subsec, sfdu_size );
                  }
                  
                  break;
               
               case GIM_NORMAL_APID:
               
                  /*
                  GIM normal is a split packet. Since the packet was
                  split into two parts, handle the data block
                  depending on whether it's first or last chuck.
                  */
                  switch ( segment_type )
                  {
                     case FIRST:
                        
                        /*
                        This is the first chunk. We copy it to the 
                        buffer that will hold assembled packet.
                        */
                        memmove( gim_norm_sfdu, sfdu, sfdu_size );
                        gim_norm_size   = sfdu_size;
                        gim_norm_sec    = sfdu_sec;
                        gim_norm_subsec = sfdu_subsec;
                        break;
                                                
                     case LAST:
                     
                        /*
                        This is the second chunk. If it has the same 
                        time, we copy it to the buffer that will hold 
                        assembled packet. Otherwise we report two 
                        "orphan" packets.
                        */
                        if ( gim_norm_sec    == sfdu_sec    && 
                             gim_norm_subsec == sfdu_subsec    )
                        {
                           memmove( &gim_norm_sfdu[gim_norm_size], 
                                    sfdu, 
                                    sfdu_size );
                           gim_norm_size += sfdu_size;
                           GIM_NORMAL_assembled = 1;
                        
                           printf( "accumulated GIM normal size %d\n", 
                                    gim_norm_size);
                        }
                        else 
                        {
                           printf( 
                              "WARNING: dropped orphan GIM_Norm FIRST @%10d:%03d\n",
                              gim_norm_sec, gim_norm_subsec );
                           printf( 
                              "WARNING: dropped orphan GIM_Norm LAST  @%10d:%03d\n",
                              sfdu_sec, sfdu_subsec );
                                  
                        }

                        break;
                                                
                  };
                  
                  /*
                  Check whether we have collected all pieces of one 
                  packet.
                  */
                  if ( GIM_NORMAL_assembled )
                  {
                  
                     /*
                     Check packet size. If OK, process the packet.
                     */
                     if ( gim_norm_size == GIM_NORMAL_SIZE ) 
                     {

                        /*
                        Initialize GIM normal HDF structures if needed.
                        */
                        if ( ! GIM_NORMAL_started )
                        {
                           if ( init_wr_gim_normal ( hdf_fp_out, 
                                             hdf_sd_out, 
                                             hdf_an_out, 
                                             "tlm_edr_data" ) == FAIL )
                           {
                              printf(  
                                 "ERROR: init_wr_gim_normal failed\n");
                              exit(-1);
                           } 
                           else 
                           {
                              GIM_NORMAL_started = 1;
                           
                              printf( "   initialized GIM Normal\n" );
                           }
                        }
                        
                        /*
                        Package and write current data block.
                        */
                        put_GIM_normal ( gim_norm_sfdu );
                     }
                     else
                     {
                        printf( 
                           "WARNING: dropped GIM Normal @%10d:%03d -- bad size %d\n",
                           sfdu_sec, sfdu_subsec, gim_norm_size );
                     }
                                             
                     /*
                     Current GEM normal packet was processed.
                     */
                     GIM_NORMAL_assembled = 0;
                  }
                  
                  break;
               
               case GIM_DIAG_APID:
               
                  /*
                  GIM diagnostic is a split packet. Handle the data
                  block depending on whether it's first, continuation
                  or last segment.
                  */
                  switch ( segment_type )
                  {
                     case FIRST:
                        
                        /*
                        This is the first chunk. We copy it to the 
                        buffer that holds assembled packet.
                        */
                        memmove( gim_diag_sfdu, sfdu, sfdu_size );
                        gim_diag_size   = sfdu_size;
                        gim_diag_sec    = sfdu_sec;
                        gim_diag_subsec = sfdu_subsec;
                        gim_diag_seqcnt = sfdu_seqcnt;
                        break;
                        
                     case CONTINUATION:
                     
                        /*
                        This is a middle chunk. If it has the same 
                        time and correct sequence count, i.e. the 
                        count of this packet is the count of previous 
                        +1 or 0, we copy it to the buffer that holds 
                        assembled packet. Otherwise we report "orphan"
                        packet.
                        */
                        if (   gim_diag_sec    == sfdu_sec    && 
                               gim_diag_subsec == sfdu_subsec &&
                             ( sfdu_seqcnt == (gim_diag_seqcnt+1) ||
                               sfdu_seqcnt == 0                      ) )
                        {
                           memmove( &gim_diag_sfdu[gim_diag_size], 
                                    sfdu, 
                                    sfdu_size );
                           gim_diag_size += sfdu_size;
                           gim_diag_seqcnt = sfdu_seqcnt;
                        } 
                        else 
                        {
                           printf( 
                              "WARNING: dropped orphan GIM_Diag MIDDLE @%10d:%03d\n",
                              sfdu_sec, sfdu_subsec );
                        }
                        break;
                        
                     case LAST:
                     
                        /*
                        This is the last chunk. If it has the same 
                        time and  correct sequence count, i.e. the 
                        count of this packet is the count of previous 
                        +1 or 0 in case of rollover, we copy it to 
                        the buffer that holds assembled packet. 
                        Otherwise we report "orphan" packets.
                        */
                        if (   gim_diag_sec    == sfdu_sec    && 
                               gim_diag_subsec == sfdu_subsec &&
                             ( sfdu_seqcnt == (gim_diag_seqcnt+1) ||
                               sfdu_seqcnt == 0                      ) )
                        {
                           memmove( &gim_diag_sfdu[gim_diag_size], 
                                    sfdu, 
                                    sfdu_size );
                           gim_diag_size += sfdu_size;
                           GIM_DIAG_assembled = 1;
                        
                           printf( "accumulated GIM diag. size %d\n", 
                                    gim_diag_size );
                        }
                        else 
                        {
                           printf( 
                              "WARNING: dropped orphan GIM_Diag FIRST @%10d:%03d\n",
                              gim_diag_sec, gim_diag_subsec );
                           printf( 
                              "WARNING: dropped orphan GIM_Diag LAST  @%10d:%03d\n",
                              sfdu_sec, sfdu_subsec );
                        }
                        break;
                        
                  };
                  
                  /*
                  Check whether we have collected all pieces of one 
                  packet.
                  */
                  if ( GIM_DIAG_assembled )
                  {
                     /*
                     Check packet size. If OK, process the packet.
                     */
                     if ( gim_diag_size == GIM_DIAG_SIZE ) 
                     {
                        /*
                        Initialize GIM diagnostics HDF structures if
                        needed.
                        */
                        if ( ! GIM_DIAG_started )
                        {
                           if ( init_wr_gim_diag ( hdf_fp_out, 
                                             hdf_sd_out, 
                                             hdf_an_out, 
                                             "tlm_edr_data" ) == FAIL )
                           {
                              printf(  
                                 "ERROR: init_wr_gim_diag failed\n");
                              exit(-1);
                           } 
                           else 
                           {
                              GIM_DIAG_started = 1;
                           
                              printf( "   initialized GIM Diag.\n" );
                           }
                        }
                        
                        /*
                        Package and write current data block.
                        */
                        put_GIM_diag ( gim_diag_sfdu );
                     }
                     else
                     {
                        printf( 
                           "WARNING: dropped GIM Diag @%10d:%03d -- bad  size %d\n",
                           sfdu_sec, sfdu_subsec, gim_diag_size );
                     }
                                             
                     /*
                     Current GIM diagnostics packet was processed.
                     */
                     GIM_DIAG_assembled = 0;
                  }
                  break;

               case GIM_RAW_APID:
               
                  /*
                  GIM raw is a split packet. Handle the data
                  block depending on whether it's first, continuation
                  or last segment.
                  */
                  switch ( segment_type )
                  {
                     case FIRST:
                        
                        /*
                        This is the first chunk. We copy it to the 
                        buffer that holds assembled packet.
                        */
                        memmove( gim_raw_sfdu, sfdu, sfdu_size );
                        gim_raw_size   = sfdu_size;
                        gim_raw_sec    = sfdu_sec;
                        gim_raw_subsec = sfdu_subsec;
                        gim_raw_seqcnt = sfdu_seqcnt;
                        break;
                        
                     case CONTINUATION:
                     
                        /*
                        This is a middle chunk. If it has the same 
                        time and correct sequence count, i.e. the 
                        count of this packet is the count of previous 
                        +1 or 0, we copy it to the buffer that holds 
                        assembled packet. Otherwise we report "orphan"
                        packet.
                        */
                        if (   gim_raw_sec    == sfdu_sec    && 
                               gim_raw_subsec == sfdu_subsec &&
                             ( sfdu_seqcnt == (gim_raw_seqcnt+1) ||
                               sfdu_seqcnt == 0                      ) )
                        {
                           memmove( &gim_raw_sfdu[gim_raw_size], 
                                    sfdu, 
                                    sfdu_size );
                           gim_raw_size += sfdu_size;
                           gim_raw_seqcnt = sfdu_seqcnt;
                        } 
                        else 
                        {
                           printf( 
                              "WARNING: dropped orphan GIM_Raw MIDDLE @%10d:%03d\n",
                              sfdu_sec, sfdu_subsec );
                        }
                        break;
                        
                     case LAST:
                     
                        /*
                        This is the last chunk. If it has the same 
                        time and  correct sequence count, i.e. the 
                        count of this packet is the count of previous 
                        +1 or 0 in case of rollover, we copy it to 
                        the buffer that holds assembled packet. 
                        Otherwise we report "orphan" packets.
                        */
                        if (   gim_raw_sec    == sfdu_sec    && 
                               gim_raw_subsec == sfdu_subsec &&
                             ( sfdu_seqcnt == (gim_raw_seqcnt+1) ||
                               sfdu_seqcnt == 0                      ) )
                        {
                           memmove( &gim_raw_sfdu[gim_raw_size], 
                                    sfdu, 
                                    sfdu_size );
                           gim_raw_size += sfdu_size;
                           GIM_RAW_assembled = 1;
                        
                           printf( "accumulated GIM raw size %d\n", 
                                    gim_raw_size );
                        }
                        else 
                        {
                           printf( 
                              "WARNING: dropped orphan GIM_Raw FIRST @%10d:%03d\n",
                              gim_raw_sec, gim_raw_subsec );
                           printf( 
                              "WARNING: dropped orphan GIM_Raw LAST  @%10d:%03d\n",
                              sfdu_sec, sfdu_subsec );
                        }
                        break;
                        
                  };
                  
                  /*
                  Check whether we have collected all pieces of one 
                  packet.
                  */
                  if ( GIM_RAW_assembled )
                  {
                     /*
                     Check packet size. If OK, process the packet.
                     */
                     if ( gim_raw_size == GIM_RAW_SIZE ) 
                     {
                        /*
                        Initialize GIM raw HDF structures if
                        needed.
                        */
                        if ( ! GIM_RAW_started )
                        {
                           if ( init_wr_gim_raw ( hdf_fp_out, 
                                             hdf_sd_out, 
                                             hdf_an_out, 
                                             "tlm_edr_data" ) == FAIL )
                           {
                              printf(  
                                 "ERROR: init_wr_gim_raw failed\n");
                              exit(-1);
                           } 
                           else 
                           {
                              GIM_RAW_started = 1;
                           
                              printf( "   initialized GIM Raw\n" );
                           }
                        }
                        
                        /*
                        Package and write current data block.
                        */
                        put_GIM_raw ( gim_raw_sfdu );
                     }
                     else
                     {
                        printf( 
                           "WARNING: dropped GIM Raw @%10d:%03d -- bad  size %d\n",
                           sfdu_sec, sfdu_subsec, gim_raw_size );
                     }
                                             
                     /*
                     Current GIM raw packet was processed.
                     */
                     GIM_RAW_assembled = 0;
                  }
                  break;
                  
            };
            break;
   
         /*
         Process concentrator APIDs.
         */
         case CON:
   
            switch ( apid_ID )
            {
         
               case CON_HK_APID:
   
                  /*
                  Check packet size -- must be greater than zero and
                  multiple of elementary data chunks. If OK, process 
                  the packet.
                  */
                  if ( sfdu_size > 0                && 
                       sfdu_size % CON_HK_SIZE == 0     ) 
                  {
                     /*
                     Initialize CON h/k HDF structures if needed.
                     */
                     if ( ! CON_HK_started )
                     {
                        if ( init_wr_con_hk ( hdf_fp_out, 
                                              hdf_sd_out, 
                                              hdf_an_out, 
                                              "tlm_edr_data" ) == FAIL )
                        {
                           printf(  
                              "ERROR: init_wr_con_hk failed\n");
                           exit(-1);
                        } 
                        else 
                        {
                           CON_HK_started = 1;
                        
                           printf( "   initialized Concentrator h/k\n");
                        }
                     }
                     
                     /*
                     Concentrator h/k is a sol packet but it can contain 
                     multiple data chunks -- call function that 
                     packages each into a structure and writes it  
                     to the file.
                     */
                     for ( i = 0; i < sfdu_size / CON_HK_SIZE; i++ ) 
                     {
                        put_CON_hk( &sfdu[0 + i * CON_HK_SIZE] );
                     }
                     
                     /*
                     Current Concentrator h/k packet has been processed.
                     */
                  }
                  else
                  {
                     printf( 
                        "WARNING: dropped CON HK @%10d:%03d -- bad size %d\n",
                        sfdu_sec, sfdu_subsec, sfdu_size );
                  }
                  
                  break;
            
            };
            break;
               
         /*
         Process MEC data.
         */
         case MEC:
   
            switch ( apid_ID )
            {
         
               case MEC_HK_APID:
               
                  /*
                  Check packet size -- must be greater than zero and
                  multiple of elementary data chunks. If OK, process 
                  the packet.
                  */
                  if ( sfdu_size > 0                && 
                       sfdu_size % MEC_HK_SIZE == 0     ) 
                  {
                     /*
                     Initialize MEC h/k HDF structures if needed.
                     */
                     if ( ! MEC_HK_started )
                     {
                        if ( init_wr_mec_hk ( hdf_fp_out, 
                                             hdf_sd_out, 
                                             hdf_an_out, 
                                             "tlm_edr_data" ) == FAIL )
                        {
                           printf(  
                                    "ERROR: init_wr_mec_hk failed\n");
                           exit(-1);
                        } 
                        else 
                        {
                           MEC_HK_started = 1;
   
                           printf( "   initialized MEC/WIND h/k\n" );
                        }
                     }
                     
                     /*
                     MEC/WIND h/k is a sol packet but it can contain 
                     multiple data chunks -- call function that 
                     packages each into a structure and writes it  
                     to the file.
                     */
                     for ( i = 0; i < sfdu_size / MEC_HK_SIZE; i++ ) 
                     {
                        put_MEC_hk( &sfdu[0 + i * MEC_HK_SIZE] );
                     }
      
                     /*
                     Current MEC h/k  packet was processed.
                     */
                  }
                  else
                  {
                     printf( 
                        "WARNING: dropped MEC HK @%10d:%03d -- bad size %d\n",
                        sfdu_sec, sfdu_subsec, sfdu_size );
                  }
                  
                  break;
                  
               case MEC_F1D_APID:
               
                  /*
                  Check packet size -- must be greater than zero and
                  multiple of elementary data chunks. If OK, process 
                  the packet.
                  */
                  if ( sfdu_size > 0                && 
                       sfdu_size % MEC_F1D_SIZE == 0     ) 
                  {
                     /*
                     Initialize MEC F1D HDF structures if needed.
                     */
                     if ( ! MEC_F1D_started )
                     {
                        if ( init_wr_mec_f1d ( hdf_fp_out, 
                                             hdf_sd_out, 
                                             hdf_an_out, 
                                             "tlm_edr_data" ) == FAIL )
                        {
                           printf(  
                                    "ERROR: init_wr_mec_f1d failed\n");
                           exit(-1);
                        } 
                        else 
                        {
                           MEC_F1D_started = 1;
   
                           printf( "   initialized MEC/WIND F1D\n" );
                        }
                     }
                     
                     /*
                     MEC/WIND F1D is a sol packet but it can contain 
                     multiple data chunks -- call function that 
                     packages each into a structure and writes it  
                     to the file.
                     */
                     for ( i = 0; i < sfdu_size / MEC_F1D_SIZE; i++ ) 
                     {
                        put_MEC_f1d( &sfdu[0 + i * MEC_F1D_SIZE] );
                     }
      
                     /*
                     Current WIND/MEC F1D  packet was processed.
                     */
                  }
                  else
                  {
                     printf( 
                        "WARNING: dropped MEC F1D @%10d:%03d -- bad size %d\n",
                        sfdu_sec, sfdu_subsec, sfdu_size );
                  }
                  
                  break;
            };
            
            break;
            
         /*
         Process SPIN TRACK quaternion APIDs.
         */
         case ACS:
   
            switch ( apid_ID )
            {
         
               case STRK_QUAT_APID:

                  /*
                  Check whether output file has been opened. This is 
                  a really redundant check, but we will do it anyway.
                  */
                  if ( TXT_opened == 0 )
                  {
                     printf( "ERROR: output file has not been opened.\n" );
                     exit(-1);
                  }

                  /*
                  Check packet size -- must be greater than zero and
                  multiple of elementary data chunks. If OK, process 
                  the packet.
                  */
                  if ( sfdu_size > 0                && 
                       sfdu_size % STRK_QUAT_SIZE == 0     ) 
                  {  
                     /*
                     Spin track quaternions APID is a sol packet but 
                     it can contain multiple data chunks. Write each 
                     of the "chunks" into output text file.
                     */
                     for ( i = 0; i < sfdu_size / STRK_QUAT_SIZE; i++ ) 
                     {

                        /*
                        Set offset for current quaternion.
                        */
                        quatoff = i * STRK_QUAT_SIZE;

                        /*
                        "decom" quaternion components.
                        */
                        C64_TO_F64( &sfdu[quatoff],      quat[0] );
                        C64_TO_F64( &sfdu[quatoff +  8], quat[1] );
                        C64_TO_F64( &sfdu[quatoff + 16], quat[2] );
                        C64_TO_F64( &sfdu[quatoff + 24], quat[3] );
                        
                        /*
                        "decom" time.
                        */
                        C64_TO_F64( &sfdu[quatoff + 32], quattim );
                        quatsub = modf( quattim, &quatsec );
                        
                        /*
                        Print quaternion record to output text file.
                        */
                        fprintf( txt_fp_out,
                           "%010.0f:%03.0f %-16.10e %-16.10e %-16.10e %-16.10e\n",
                           quatsec, quatsub*256.0, 
                           quat[0], quat[1], quat[2], quat[3] );
                        
                     }
                     
                     /*
                     Current Spin track quaternions packet has been
                     processed.
                     */
                  }
                  else
                  {
                     printf( 
                        "WARNING: dropped STRK_QUAT @%10d:%03d -- bad size %d\n",
                        sfdu_sec, sfdu_subsec, sfdu_size );
                  }
                  
                  break;
            
            };
            break;
               
      }
   
   }
   
   /*
   Yahoo! we have completed this run. Time to wrap up whatever 
   HDF structures we have initialized and close the file.
   */
   if ( HDF_opened )
   {
      /*
      Let's see which HDF structures we started and close them all,
      one at a time. But, just in case, let verify that we have 
      started at least one.
      */
      if (  GEM_HK_started       || 
            GEM_MANCAL_started   || 
            GEM_NORMAL_started   || 
            GEM_RAW_started      || 
            GIM_HK_started       || 
            GIM_MANCAL_started   || 
            GIM_NORMAL_started   || 
            GIM_DIAG_started     || 
            GIM_RAW_started      || 
            CON_HK_started       || 
            MEC_HK_started       || 
            MEC_F1D_started        )
      {
         if ( GEM_HK_started )      close_wr_gem_hk();
         if ( GEM_MANCAL_started )  close_wr_gem_mancal();
         if ( GEM_NORMAL_started )  close_wr_gem_normal();
         if ( GEM_RAW_started )     close_wr_gem_raw();
         if ( GIM_HK_started )      close_wr_gim_hk();
         if ( GIM_MANCAL_started )  close_wr_gim_mancal();
         if ( GIM_NORMAL_started )  close_wr_gim_normal();
         if ( GIM_DIAG_started )    close_wr_gim_diag();
         if ( GIM_RAW_started )     close_wr_gim_raw();
         if ( CON_HK_started )      close_wr_con_hk();
         if ( MEC_HK_started )      close_wr_mec_hk();
         if ( MEC_F1D_started )     close_wr_mec_f1d();
      }
      else
      {
         /*
         This is really really strange -- we have opened the file
         but didn't start any structures. This could have happened 
         only if we have had some number of chucnks of long packets 
         but never managed to assemble the whole packet from them. 
         In any case the output file is worthless and user should 
         know about it.
         */
         printf(  "ERROR: Output EDR file is empty.\n" );
         exit(-1);
      }
      
      /*
      Now, when we wrapped up all opened structures, we can close
      HDF Vdata and SD "areas" and close HDF file. 
      */
      if ( SDend ( hdf_sd_out ) == FAIL )
      {
         printf(  "ERROR: SDend failed\n" );
         exit(-1);
      }
      
      if ( Vend  ( hdf_fp_out ) == FAIL )
      {
         printf(  "ERROR: Vend failed\n" );
         exit(-1);
      }
      
      if ( Hclose( hdf_fp_out ) == FAIL )
      {
         printf(  "ERROR: Hclose failed\n" );
         exit(-1);
      }
      
      /*
      If we are here, then the run was a complete success. The only 
      thing left to do is to rename temporary file to be actual HDF 
      file.
      */
      if ( rename( tmp_file, hdf_file ) == 0 )
      {
         printf( "Program run completed successfully.\n" );
         exit(0);
      }
      else
      {
         printf(  "ERROR: Failed to rename temporary file.\n" );
         exit(-1);
      }

   }
   else if ( TXT_opened )
   {
      /*
      Close output text file with quaternions.
      */
      
      if ( fclose( txt_fp_out ) ) 
      {
         printf( "ERROR: Failed to close output file %s.\n", tmp_file );
         exit(-1);
      }
   
      /*
      If we are here, then the run was a complete success. The only 
      thing left to do is to rename temporary file to be actual text 
      file.
      */
      if ( rename( tmp_file, hdf_file ) == 0 )
      {
         printf( "Program run completed successfully.\n" );
         exit(0);
      }
      else
      {
         printf( "ERROR: Failed to rename temporary file.\n" );
         exit(-1);
      }
      
   }
   else
   {
      /*
      Well, it looks like we have never opened output file. Let's 
      make sure that user knows about it.
      */
      printf(  "ERROR: Output EDR or text file was not created.\n" );
      exit(-1);
   }

}

/**********************************************************************/
/**********************************************************************/
/**********************************************************************/

/*
This function is heavily based on the Robin O'Brien's "generic.c" and 
Howard Taylor's "dmapktdecom".
*/

int got_good_apid( char  * tlm_file,
                   int   * apid_ID,
                   int   * segment_type,
                   int   * sfdu_size,
                   int   * sfdu_sec,
                   int   * sfdu_subsec,
                   int   * sfdu_seqcnt,
                   uint8 * sfdu   )
{
   /*
   Input TLM file descriptor
   */
   static int tlm_input = 0;
   
   /*
   Input file opened flag and "first time here" flags (the second of 
   these flags isn't really necessary, but we will keep for sake of 
   similarity with the original code.) 
   */
   static int tlm_file_opened = 0;
   static int first_time = 1;
   
   /*
   Misc. counters.
   */
   static int accumulated_byte_count = 0;
   static int sfdu_count = 1;
   
   /*
   Pointers to various parts of the complete SFDU packet.
   */
   uint8 * ccsds_ptr;
   uint8 * sclk_ptr;
   uint8 * data_ptr;
   uint8 * byte_ptr;
   
   /*
   Sizes of various parts of the complete SFDU packet.
   */
   int sfdu_len;
   int header_len;
   int data_len;
   int len;
   
   /*
   Misc. variables.
   */
   int check;

   /*
   These ones are from Howie's "libdma/getpktinfo.c". They are 
   pointers to various sub-headers within the main SFDU aggregate 
   header.
   */
   uint8 * sfdu_label;
   uint8 * sfdu_aggregate;
   uint8 * sfdu_primary;
   uint8 * sfdu_secondary;
   uint8 * sfdu_tertiary;
   
   /*
   These ones came from Howie's ApidHeader "libdma/dma.h". They
   are packet field flag and non-filled lenght from the tertiary
   header.
   */
   unsigned char  pkt_fill_flag;
   unsigned short non_fill_length;

   /*
   These ones came from Howie's ApidHeader "libdma/dma.h". They
   are fields of the data CHDO header.
   */
   unsigned char  header_version; /* should be 000 */
   unsigned char  header_type;    /* should be   0 */
   unsigned char  secondary_type; /* should be   1 */
   unsigned short sequence_count; /* should increment */
   unsigned short pktlen;         /* packet length */
   unsigned int   sclk_sec;       /* packet sclk seconds */
   unsigned char  sclk_sub;       /* packet sclk subseconds */
   
   /*
   If TLM file is not opened yet, open it.
   */
   if ( ! tlm_file_opened )
   {
      tlm_input = open( tlm_file, O_RDONLY, 0 );
      if ( tlm_input < 0 )
      {
         printf(  "ERROR: Failed to open TLM file.\n" );
         exit(-1);
      }
      else
      {
         tlm_file_opened = 1;
      }
   }
   
   /*
   Read next SFDU header from the file.
   */
   len = read( tlm_input, sfdu, HEADER );
   if ( len > 0 )
   {
      /*
      If we have got less bytes that we were suppose to get, signal
      an error.
      */
      if ( len != HEADER )
      {
         printf(  "ERROR: Bad SFDU Header, Rec # %d\n", 
                  sfdu_count );
         close  ( tlm_input );
         exit(-1);
      }
      
      /*
      Checks the first SFDU in the file. All valid SFDUs should 
      begin with the following 32 bit string:
      */
      if ( first_time )
      {
         check = C32_TO_U32( sfdu );

         if ( check != 0x4e4a504c )
         {
            printf(  "ERROR: File starts with invalid sfdu.\n");
            close  ( tlm_input );
            exit(-1);
         }
         
         first_time = 0;
      }

      /*
      Move pointer to the 19th byte of the header and retrieve 
      block lenght from the 19th and 20th bytes.
      */
      byte_ptr = sfdu + 18;
      sfdu_len = C16_TO_U16(byte_ptr);
      
      /*
      Read the whole block into the buffer.
      */
      len = read( tlm_input, sfdu+len, sfdu_len );
      
      /*
      Check whether the number of bytes that we got in the same as 
      the number that was in the SFDU header.
      */
      if ( len != sfdu_len )
      {
         printf(  "ERROR: Bad SFDU Length %d, Rec # %d\n", 
                  sfdu_len, sfdu_count );
         close  ( tlm_input );
         exit(-1);
      }

      /*
      Add SFDU header lenght to the data block lenght to get total
      SFDU packet lenght.
      */
      sfdu_len += HEADER;
      
      /*
      At this point we have the whole SFDU packet (with all headers,
      etc.) in the memory pointed by "sfdu". The size of that memory 
      block is in the "sfdu_len". The next step is to determine where 
      the actual data starts. 
      */
      
      /* 
      Calculate offset for aggregation CHDO chdo_length 
      */
      byte_ptr = sfdu + 22;
      
      /*
      Calculate total SFDU header lenght.
      */
      header_len  = C16_TO_U16(byte_ptr) + HEADER + AGGREGATE + 
                    DATA_CHDO;
   
      /*
      Calculate the length of the data area ONLY of the sfdu.
      */
      data_len = sfdu_len - header_len;
      
      /*
      Set all pointers -- for packet primary CCSDS header, packet
      creation sclk, beginning of the data and check sum.
      */
      ccsds_ptr  = sfdu + header_len;
      sclk_ptr   = ccsds_ptr + 6;
      data_ptr   = sclk_ptr + 5;
      
      /*
      Decom primary and secondary packet headers (this code was 
      borrowed from Howie's "libdma/unpackapidhdr.c".)
      */
      header_version = ( ccsds_ptr[0] >> 5 ) & 0x03;
      header_type    = ( ccsds_ptr[0] >> 4 ) & 0x01;
      secondary_type = ( ccsds_ptr[0] >> 3 ) & 0x01;
      *apid_ID       = C16_TO_U16 ( ccsds_ptr ) & 0x07ff;
      *segment_type  = ( ccsds_ptr[2] >> 6 ) & 0x03;
      sequence_count = C16_TO_U16 ( ccsds_ptr + 2 ) & 0x3fff;
      pktlen         = C16_TO_U16 ( ccsds_ptr + 4 ) & 0xffff;
      sclk_sec       = C32_TO_U32 ( sclk_ptr );
      sclk_sub       = sclk_ptr[4];

      /*
      Now we need to find out whether this packet is a good, i.e. fully
      filled, packet or not. Information about filling, flag and 
      original, non-filled lenght, are in the tertiary header. 
      */
      sfdu_label      = sfdu;
      sfdu_aggregate  = sfdu_label     + 20;
      sfdu_primary    = sfdu_aggregate + 4;           
      sfdu_secondary  = sfdu_primary   + 4 + 
                        C16_TO_U16( sfdu_primary   + 2 );
      sfdu_tertiary   = sfdu_secondary + 4 + 
                        C16_TO_U16( sfdu_secondary + 2 );

      pkt_fill_flag   = (sfdu_tertiary[CHDO83_PKT_FILL_FLAG] &
                        CHDO83_PKT_FILL_FLAG_MASK) >> 7;

      non_fill_length = C16_TO_U16 ( sfdu_tertiary + 
                                     CHDO83_NON_FILL_LENGTH ) & 
                        CHDO83_NON_FILL_LENGTH_MASK;
                        
      /*
      Report all essentials about this packet.
      */
      printf( "%2d %1d %6d %4d %10d:%03d (%d/%d/%d/%d) (fill:%d/%d)\n", 
              *apid_ID, 
              *segment_type, 
              sequence_count, 
              pktlen + 1 - 5, 
              sclk_sec, 
              sclk_sub, 
              sfdu_count, 
              sfdu_len, 
              header_len, 
              data_len - 6 - 5,
              pkt_fill_flag,
              non_fill_length - 6 - 5 );
              
      /*
      If packet is good (not filled), prepare the data for output. 
      Otherwise, complain and change APID ID to 0.
      */
      if ( ! pkt_fill_flag ) 
      {
         /*
         Re-assign data size and copy data to the location pointed by 
         input pointer.
         */
         *sfdu_size = pktlen + 1 - 5;
         memmove( sfdu, data_ptr, *sfdu_size );
         
         /*
         Re-assign packet time and sequence count.
         */
         *sfdu_sec    = sclk_sec;
         *sfdu_subsec = sclk_sub;
         *sfdu_seqcnt = sequence_count;
      }
      else
      {
         /*
         This packet was filled. Complain and reset APID ID to zero 
         to make sure that no processing is done on this packet by the
         caller.
         */
         printf(  
            "WARNING: dropped filled APID %2d @%10d:%03d. Orig size: %d.\n",
            *apid_ID, *sfdu_sec, *sfdu_subsec, non_fill_length );
         
         *apid_ID = 0;
      }
      
      /*
      Current SFDU was read in succesfully. Increase the counter.
      */
      sfdu_count++;
      accumulated_byte_count += sfdu_len;
      
      return( sfdu_count );
   }
   else 
   {
      /*
      If looks like we have reached the end of the file.
      */
      close  ( tlm_input );
      printf ( "Accumulated processed byte count: %d\n", 
               accumulated_byte_count );
      return ( 0 );
   }

}


/**********************************************************************/
/**********************************************************************/
/**********************************************************************/

/*
This function opens HDF file and intializes Vdata, SD and Annotation
interfaces.
*/
int open_HDF_file (  char  * hdf_file, 
                     int32 * hdf_fp_out, 
                     int32 * hdf_sd_out, 
                     int32 * hdf_an_out )
{

   /*
   All calls below were copied from Ed'd foo.c HDF driver.
   */

   /*
   Open new HDF file for write access.
   */
   if ( ( *hdf_fp_out = Hopen( hdf_file, DFACC_CREATE, 0 ) ) == FAIL )
   {
      printf(  "ERROR: Hopen failed\n" );
      perror ( "Hopen(out)" );
      exit(-1);
   }
   
   /*
   Start up HDF Vdata interface.
   */
   if ( Vstart( *hdf_fp_out ) != 0 ) 
   {
      printf(  "ERROR: Vstart failed\n" );
      exit(-1);
   }

   /*
   Start up HDF SD (Scientific Data) Interface.
   */
   if ( ( *hdf_sd_out = SDstart( hdf_file, DFACC_RDWR ) ) == FAIL )
   {
      printf(  "ERROR: SDstart failed\n" );
      exit(-1);
   }

   /*
   Start up the HDF Annotation interface
   */
   if( ( *hdf_an_out = ANstart( *hdf_fp_out ) ) == FAIL )
   {
      printf(  "ERROR: ANstart failed\n" );
      exit(-1);
   }

}

/**********************************************************************/
/**********************************************************************/
/**********************************************************************/

/*
   This function packages block of bytes representing 
   complete GEM normal APID and writes it to HDF output file.
*/

int put_GEM_normal ( uint8 * data )
{
   /*
   This is the structure that we need to fill out.
   */
   struct GEM_Normal g;
   
   /*
   These are misc counters and offsets.
   */
   int i;
   int secoff;
   int suboff;
   int stpoff;
   int lvloff;
   int sizoff;
   
   /*
   Let get to work. Note that all offsets for the input data block 
   are hard coded. They correspond to the GEM Normal APID layout
   provided by Karl Lauffer on June 19, 2000.
   */

   /* 
   First, assign packet status -- first two bytes of the data block.
   */
   g.APID_status   = U8_TO_U16( data[0], data[1] );

   /*
   Assign time -- sec and subsec, -- time step, step level and step
   size for each of four spins. These come in subsequent blocks of 12
   bytes. In each block:
      bytes  0..3  -- SCLK seconds;
      byte   4..5  -- SCLK subseconds;
      bytes  6..7  -- time step;
      bytes  8..9  -- level;
      bytes 10..11 -- step size;
   First block starts at byte 2 of the data.
   */  
   for ( i=0; i < GEM_SPINS; i++ )
   {
      /*
      Calculate offset for the first byte of each value in i-th 
      spin block and assign the strcuture values.
      */
      secoff = 2 + 12 * i;
      suboff = 2 + 12 * i + 4;
      stpoff = 2 + 12 * i + 6;
      lvloff = 2 + 12 * i + 8;
      sizoff = 2 + 12 * i + 10;

      g.sclk_seconds[i]    = U8_TO_U32(  data[secoff], data[secoff+1],
                                       data[secoff+2], data[secoff+3]);
      g.sclk_subseconds[i] = U8_TO_U16(  data[suboff], data[suboff+1]);
      g.step_time[i]       = U8_TO_U16(  data[stpoff], data[stpoff+1]);
      g.level[i]           = U8_TO_U16(  data[lvloff], data[lvloff+1]);
      g.step_size[i]       = U8_TO_U16(  data[sizoff], data[sizoff+1]);
   }
   
   /*
   Assign integration time, bytes 50..51, and settle time, bytes 52..53,
   of the data block.
   */
   g.integ_time   = U8_TO_U16( data[50], data[51] );
   g.settle_time  = U8_TO_U16( data[52], data[53] );;

   /*
   Assign ESA levels contained in the next 20 2-bytes words (starting 
   with byte 54 of the data block.)
   */
   for ( i=0; i < GEM_CYCLE_LEVELS; i++ )
   {
      /*
      Calculate offset of the next ESA level.
      */
      lvloff = 54 + 2 * i;
      g.ESA_level[i] = U8_TO_U16( data[lvloff], data[lvloff+1] );
   }
   
   /*
   And, at last, move the whole counts block -- 24*7*20 bytes starting 
   at byte 94 -- into the counts field.
   */
   memmove(&g.counts[0][0][0], &data[94], 3360);

   /*
   We finished packaging byte block into the structure and can write
   it to the file.
   */
   if ( write_gem_normal( g, -1 ) == FAIL ) {
      printf(  "ERROR: write_gem_normal() failed\n" );
      exit(-1);
   }
   printf( "      wrote GEM normal record\n" );
}

/**********************************************************************/
/**********************************************************************/
/**********************************************************************/

/*
   This function packages block of bytes representing 
   complete GEM raw count APID and writes it to HDF output file.
*/

int put_GEM_raw ( uint8 * data )
{
   /*
   This is the structure that we need to fill out.
   */
   struct GEM_Raw g;
   
   /*
   These are misc counters and offsets.
   */
   int i;
   int j;
   int k;
   int sblkoff;
   int lblkoff;
   int cblkoff;
  
   /*
   Let get to work. Note that all offsets for the input data block 
   are hard coded. They correspond to the GEM Raw APID layout
   provided by Dave Gingrich and Karl Lauffer, LMA on November 11, 2000.
   */

   /* 
   First, assign packet time -- first eight bytes of the data block
   (last two of eight are pad.)
   */
   g.sclk_seconds    = U8_TO_U32( data[0], data[1], data[2], data[3]);
   g.sclk_subseconds = U8_TO_U16( data[4], data[5]);

   /*
   Loop to process sector data blocks.
   */  
   for ( i=0; i < GEM_AZIMUTH_SECTORS; i++ )
   {

      /*
      Offset for current sector data block.
      */
      sblkoff = 8 + 98 * i;

      /*
      Assign per-sector structure values.
      */
      g.start_req[i] = U8_TO_U16( data[sblkoff  ], data[sblkoff+1]);
      g.step_size[i] = U8_TO_U16( data[sblkoff+2], data[sblkoff+3]);         
      g.step_time[i] = U8_TO_U16( data[sblkoff+4], data[sblkoff+5]);          
      g.angle[i]     = U8_TO_U16( data[sblkoff+6], data[sblkoff+7]);

      /*
      Loop to process spin levels within sector.
      */
      for ( j=0; j < GEM_SPIN_LEVELS; j++ )
      {

         /*
         Offset for current sector/level block.
         */
         lblkoff = sblkoff + 8 + 18 * j;

         /*
         Assign per-sector/level structure values.
         */
         g.ESA_set[i][j]   = U8_TO_U16( data[lblkoff  ], data[lblkoff+1]);
         g.ESA_level[i][j] = U8_TO_U16( data[lblkoff+2], data[lblkoff+3]);

         /*
         Loop to process CEM counts within sector/level.
         */
         for ( k=0; k < GEM_CEMS; k++ )
         {

            /*
            Offset for current sector/level/CEM count.
            */
            cblkoff = lblkoff + 4 + 2 * k;

            /*
            Assign sector/level/CEM count.
            */
            g.counts[i][j][k] = U8_TO_U16( data[cblkoff], data[cblkoff+1]);

         }

      }

   }
   
   /*
   We finished packaging byte block into the structure and can write
   it to the file.
   */
   if ( write_gem_raw( g, -1 ) == FAIL ) {
      printf(  "ERROR: write_gem_raw() failed\n" );
      exit(-1);
   }
   printf( "      wrote GEM raw record\n" );
}

/**********************************************************************/
/**********************************************************************/
/**********************************************************************/

/*
   This function packages block of bytes representing 
   complete GEM h/k APID and writes it to HDF output file.
*/

int put_GEM_hk ( uint8 * data )
{
   /*
   This is the structure that we need to fill out.
   */
   struct GEM_Housekeeping g;
   
   /*
   These are misc counters and offsets.
   */
   int i;
   int setoff;
   int lvloff;
   
   /*
   Let get to work. Note, that all offsets for the input data block 
   are hard coded. They correspond to the GEM h/k APID layout
   provided by Karl Lauffer on June 19, 2000.
   */

   /* 
   First, assign time  -- first 6 bytes the data block.
   */
   g.sclk_seconds    = U8_TO_U32( data[0], data[1], data[2], data[3] );
   g.sclk_subseconds = U8_TO_U16( data[4], data[5] );

   /* 
   Assign packet status -- bytes 6 and 7 of the data block.
   */
   g.APID_status     = U8_TO_U16( data[6], data[7] );
   
   /* 
   Assign start level -- bytes 8 and 9, -- and step size -- bytes 10
   and 11.
   */
   g.start_level     = U8_TO_U16( data[8], data[9] );
   g.step_size       = U8_TO_U16( data[10], data[11] );

   /*
   Assign set and level for each of five levels during one spin. These 
   come in subsequent blocks of 4 bytes. In each block:
      bytes  0..1  -- set;
      byte   2..3  -- level;
   First block starts at byte 12 of the data.
   */  
   for ( i=0; i < GEM_SPIN_LEVELS; i++ )
   {
      /*
      Calculate offset for the first byte of each value in i-th 
      level and assign the structure values.
      */
      setoff = 12 + 4 * i;
      lvloff = 12 + 4 * i + 2;

      g.set[i]    = U8_TO_U16(  data[setoff], data[setoff+1]);
      g.level[i]  = U8_TO_U16(  data[lvloff], data[lvloff+1]);
   }
   
   /*
   Assign CEM_set, CEM_level, AMP_set, MON_5V, MON_12V, SW_status and  
   CMD_status. These come as two byte words starting with byte 32 of
   the data block in the order they defined in the structure.
   */
   g.CEM_set      = U8_TO_U16( data[32], data[33] );
   g.CEM_level    = U8_TO_U16( data[34], data[35] );
   g.AMP_set      = U8_TO_U16( data[36], data[37] );
   g.MON_5V       = U8_TO_U16( data[38], data[39] );
   g.MON_12V      = U8_TO_U16( data[40], data[41] );
   g.SW_status    = U8_TO_U16( data[42], data[43] );
   g.CMD_status   = U8_TO_U16( data[44], data[45] );

   /*
   We finished packaging byte block into the structure and can write
   it to the file.
   */
   if ( write_gem_hk( g, -1 ) == FAIL ) {
      printf(  "ERROR: write_gem_hk() failed\n" );
      exit(-1);
   }
   printf( "      wrote GEM h/k record\n" );

}

/**********************************************************************/
/**********************************************************************/
/**********************************************************************/

/*
   This function packages block of bytes representing 
   complete GEM man/cal APID and writes it to HDF output file.
*/

int put_GEM_mancal ( uint8 * data )
{
   /*
   This is the structure that we need to fill out.
   */
   struct GEM_ManCal g;
   
   /*
   Let get to work. Note that all offsets for the input data block 
   are hard coded. They correspond to the GEM Man/Cal APID layout
   provided by Karl Lauffer on June 19, 2000.
   */

   /* 
   First, assign time  -- first 6 bytes the data block.
   */
   g.sclk_seconds    = U8_TO_U32( data[0], data[1], data[2], data[3] );
   g.sclk_subseconds = U8_TO_U16( data[4], data[5] );

   /* 
   Assign packet status -- bytes 6 and 7 of the data block.
   */
   g.APID_status  = U8_TO_U16( data[6], data[7] );
   
   /* 
   Assign step_time, integ_time and settle_time. These come as two 
   byte words starting with byte 8 of the data block in the order 
   they defined in the structure.
   */
   g.step_time    = U8_TO_U16( data[8], data[9] );
   g.integ_time   = U8_TO_U16( data[10], data[11] );
   g.settle_time  = U8_TO_U16( data[12], data[13] );

   /*
   And, at last, move the whole counts block -- 7*5 bytes starting 
   at byte 14 -- into the counts field.
   */
   memmove( &g.counts[0][0], &data[14], 35 );

   /*
   We finished packaging byte block into the structure and can write
   it to the file.
   */
   if ( write_gem_mancal( g, -1 ) == FAIL ) {
      printf(  "ERROR: write_gem_mancal() failed\n" );
      exit(-1);
   }
   printf( "      wrote GEM man/cal record\n" );
}

/**********************************************************************/
/**********************************************************************/
/**********************************************************************/


/*
   This function packages block of bytes representing 
   complete GIM normal APID and writes it to HDF output file.
*/

int put_GIM_normal ( uint8 * data )
{
   /*
   This is the structure that we need to fill out.
   */
   struct GIM_Normal g;
   
   /*
   These are misc counters and offsets.
   */
   int i;
   int secoff;
   int suboff;
   int stpoff;
   int lvloff;
   int sizoff;
   
   /*
   Let get to work. Note that all offsets for the input data block 
   are hard coded. They correspond to the GIM Normal APID layout
   provided by Karl Lauffer on June 19, 2000.
   */

   /* 
   First, assign packet status -- first two bytes of the data block.
   */
   g.APID_status   = U8_TO_U16( data[0], data[1] );

   /*
   Assign time -- sec and subsec, -- time step, step level and step
   size for each of four spins. These come in subsequent blocks of 12
   bytes. In each block:
      bytes  0..3  -- SCLK seconds;
      byte   4..5  -- SCLK subseconds;
      bytes  6..7  -- time step;
      bytes  8..9  -- level;
      bytes 10..11 -- step size;
   First block starts at byte 2 of the data.
   */  
   for ( i=0; i < GIM_SPINS; i++ )
   {
      /*
      Calculate offset for the first byte of each value in i-th 
      spin block and assign the strcuture values.
      */
      secoff = 2 + 12 * i;
      suboff = 2 + 12 * i + 4;
      stpoff = 2 + 12 * i + 6;
      lvloff = 2 + 12 * i + 8;
      sizoff = 2 + 12 * i + 10;

      g.sclk_seconds[i]    = U8_TO_U32(  data[secoff], data[secoff+1],
                                       data[secoff+2], data[secoff+3]);
      g.sclk_subseconds[i] = U8_TO_U16(  data[suboff], data[suboff+1]);
      g.step_time[i]       = U8_TO_U16(  data[stpoff], data[stpoff+1]);
      g.level[i]           = U8_TO_U16(  data[lvloff], data[lvloff+1]);
      g.step_size[i]       = U8_TO_U16(  data[sizoff], data[sizoff+1]);
   }
   
   /*
   Assign integration time, bytes 50..51, and settle time, bytes 52..53,
   of the data block.
   */
   g.integ_time   = U8_TO_U16( data[50], data[51] );
   g.settle_time  = U8_TO_U16( data[52], data[53] );;

   /*
   Assign ESA levels contained in the next 40 2-bytes words (starting 
   with byte 54 of the data block.)
   */
   for ( i=0; i < GIM_CYCLE_LEVELS; i++ )
   {
      /*
      Calculate offset of the next ESA level.
      */
      lvloff = 54 + 2 * i;
      g.ESA_level[i] = U8_TO_U16( data[lvloff], data[lvloff+1] );
   }
   
   /*
   Assign mask table number and azimunth with peak. These are one byte 
   values immediately following ESA levels.
   */
   g.mask_table  = data[134];
   g.peak_sector = data[135];
   
   /*
   And, at last, move the whole masked counts block -- 80*40 bytes 
   starting at byte 136 -- into the counts field.
   */
   memmove( &g.counts[0][0], &data[136], 3200 );

   /*
   We finished packaging byte block into the structure and can write
   it to the file.
   */
   if ( write_gim_normal( g, -1 ) == FAIL ) {
      printf(  "ERROR: write_gim_normal() failed\n" );
      exit(-1);
   }
   printf( "      wrote GIM normal record\n" );
}

/**********************************************************************/
/**********************************************************************/
/**********************************************************************/

/*
   This function packages block of bytes representing 
   complete GIM raw count APID and writes it to HDF output file.
*/

int put_GIM_raw ( uint8 * data )
{
   /*
   This is the structure that we need to fill out.
   */
   struct GIM_Raw g;
   
   /*
   These are misc counters and offsets.
   */
   int i;
   int j;
   int k;
   int sblkoff;
   int lblkoff;
   int cblkoff;
  
   /*
   Let get to work. Note that all offsets for the input data block 
   are hard coded. They correspond to the GIM Raw APID layout
   provided by Dave Gingrich and Karl Lauffer, LMA on November 11, 2000.
   */

   /* 
   First, assign packet time -- first six bytes of the data block
   (last two of eight are pad.)
   */
   g.sclk_seconds    = U8_TO_U32( data[0], data[1], data[2], data[3]);
   g.sclk_subseconds = U8_TO_U16( data[4], data[5]);

   /*
   Loop to process sector data blocks.
   */  
   for ( i=0; i < GIM_AZIMUTH_SECTORS; i++ )
   {

      /*
      Offset for current sector data block.
      */
      sblkoff = 8 + 208 * i;

      /*
      Assign per-sector structure values.
      */
      g.start_req[i] = U8_TO_U16( data[sblkoff  ], data[sblkoff+1]);
      g.step_size[i] = U8_TO_U16( data[sblkoff+2], data[sblkoff+3]);         
      g.step_time[i] = U8_TO_U16( data[sblkoff+4], data[sblkoff+5]);          
      g.angle[i]     = U8_TO_U16( data[sblkoff+6], data[sblkoff+7]);

      /*
      Loop to process spin levels within sector.
      */
      for ( j=0; j < GIM_SPIN_LEVELS; j++ )
      {

         /*
         Offset for current sector/level block.
         */
         lblkoff = sblkoff + 8 + 20 * j;

         /*
         Assign per-sector/level structure values.
         */
         g.ESA_set[i][j]   = U8_TO_U16( data[lblkoff  ], data[lblkoff+1]);
         g.ESA_level[i][j] = U8_TO_U16( data[lblkoff+2], data[lblkoff+3]);

         /*
         Loop to process CEM counts within sector/level.
         */
         for ( k=0; k < GIM_CEMS; k++ )
         {

            /*
            Offset for current sector/level/CEM count.
            */
            cblkoff = lblkoff + 4 + 2 * k;

            /*
            Assign sector/level/CEM count.
            */
            g.counts[i][j][k] = U8_TO_U16( data[cblkoff], data[cblkoff+1]);

         }

      }

   }
   
   /*
   We finished packaging byte block into the structure and can write
   it to the file.
   */
   if ( write_gim_raw( g, -1 ) == FAIL ) {
      printf(  "ERROR: write_gim_raw() failed\n" );
      exit(-1);
   }
   printf( "      wrote GIM raw record\n" );
}

/**********************************************************************/
/**********************************************************************/
/**********************************************************************/

/*
   This function packages block of bytes representing 
   complete GIM diagnostics APID and writes it to HDF output file.
*/

int put_GIM_diag ( uint8 * data )
{
   /*
   This is the structure that we need to fill out.
   */
   struct GIM_Diagnostic g;
   
   /*
   These are misc counters and offsets.
   */
   int i;
   int secoff;
   int suboff;
   int stpoff;
   int lvloff;
   int sizoff;
   
   /*
   Let get to work. Note that all offsets for the input data block 
   are hard coded. They correspond to the GIM Diagnostics APID layout
   provided by Karl Lauffer on June 19, 2000.
   */

   /* 
   First, assign packet status -- first two bytes of the data block.
   */
   g.APID_status   = U8_TO_U16( data[0], data[1] );

   /*
   Assign time -- sec and subsec, -- time step, step level and step
   size for each of four spins. These come in subsequent blocks of 12
   bytes. In each block:
      bytes  0..3  -- SCLK seconds;
      byte   4..5  -- SCLK subseconds;
      bytes  6..7  -- time step;
      bytes  8..9  -- level;
      bytes 10..11 -- step size;
   First block starts at byte 2 of the data.
   */  
   for ( i=0; i < GIM_SPINS; i++ )
   {
      /*
      Calculate offset for the first byte of each value in i-th 
      spin block and assign the strcuture values.
      */
      secoff = 2 + 12 * i;
      suboff = 2 + 12 * i + 4;
      stpoff = 2 + 12 * i + 6;
      lvloff = 2 + 12 * i + 8;
      sizoff = 2 + 12 * i + 10;

      g.sclk_seconds[i]    = U8_TO_U32(  data[secoff], data[secoff+1],
                                       data[secoff+2], data[secoff+3]);
      g.sclk_subseconds[i] = U8_TO_U16(  data[suboff], data[suboff+1]);
      g.step_time[i]       = U8_TO_U16(  data[stpoff], data[stpoff+1]);
      g.level[i]           = U8_TO_U16(  data[lvloff], data[lvloff+1]);
      g.step_size[i]       = U8_TO_U16(  data[sizoff], data[sizoff+1]);
   }
   
   /*
   Assign integration time, bytes 50..51, and settle time, bytes 52..53,
   of the data block.
   */
   g.integ_time   = U8_TO_U16( data[50], data[51] );
   g.settle_time  = U8_TO_U16( data[52], data[53] );;

   /*
   Assign ESA levels contained in the next 40 2-bytes words (starting 
   with byte 54 of the data block.)
   */
   for ( i=0; i < GIM_CYCLE_LEVELS; i++ )
   {
      /*
      Calculate offset of the next ESA level.
      */
      lvloff = 54 + 2 * i;
      g.ESA_level[i] = U8_TO_U16( data[lvloff], data[lvloff+1] );
   }
   
   /*
   And, at last, move the whole un-masked counts block -- 40*8*40 bytes 
   starting at byte 134 -- into the counts field.
   */
   memmove( &g.counts[0][0][0], &data[134], 12800 );

   /*
   We finished packaging byte block into the structure and can write
   it to the file.
   */
   if ( write_gim_diag( g, -1 ) == FAIL ) {
      printf(  "ERROR: write_gim_diag() failed\n" );
      exit(-1);
   }
   printf( "      wrote GIM diagnostics record\n" );
}

/**********************************************************************/
/**********************************************************************/
/**********************************************************************/

/*
   This function packages block of bytes representing 
   complete GIM man/cal APID and writes it to HDF output file.
*/

int put_GIM_mancal ( uint8 * data )
{
   /*
   This is the structure that we need to fill out.
   */
   struct GIM_ManCal g;
   
   /*
   Let get to work. Note that all offsets for the input data block 
   are hard coded. They correspond to the GIM Man/Cal APID layout
   provided by Karl Lauffer on June 19, 2000.
   */

   /* 
   First, assign time  -- first 6 bytes the data block.
   */
   g.sclk_seconds    = U8_TO_U32( data[0], data[1], data[2], data[3] );
   g.sclk_subseconds = U8_TO_U16( data[4], data[5] );

   /* 
   Assign packet status -- bytes 6 and 7 of the data block.
   */
   g.APID_status  = U8_TO_U16( data[6], data[7] );
   
   /* 
   Assign step_time, integ_time and settle_time. These come as two 
   byte words starting with byte 8 of the data block in the order 
   they defined in the structure.
   */
   g.step_time    = U8_TO_U16( data[8], data[9] );
   g.integ_time   = U8_TO_U16( data[10], data[11] );
   g.settle_time  = U8_TO_U16( data[12], data[13] );

   /*
   And, at last, move the whole counts block -- 8*10 bytes starting 
   at byte 14 -- into the counts field.
   */
   memmove( &g.counts[0][0], &data[14], 80 );

   /*
   We finished packaging byte block into the structure and can write
   it to the file.
   */
   if ( write_gim_mancal( g, -1 ) == FAIL ) {
      printf(  "ERROR: write_gim_mancal() failed\n" );
      exit(-1);
   }
   printf( "      wrote GIM man/cal record\n" );
}

/**********************************************************************/
/**********************************************************************/
/**********************************************************************/

/*
   This function packages block of bytes representing 
   complete GIM h/k APID and writes it to HDF output file.
*/

int put_GIM_hk ( uint8 * data )
{
   /*
   This is the structure that we need to fill out.
   */
   struct GIM_Housekeeping g;
   
   /*
   These are misc counters and offsets.
   */
   int i;
   int setoff;
   int lvloff;
   
   /*
   Let get to work. Note, that all offsets for the input data block 
   are hard coded. They correspond to the GIM h/k APID layout
   provided by Karl Lauffer on June 19, 2000.
   */

   /* 
   First, assign time  -- first 6 bytes the data block.
   */
   g.sclk_seconds    = U8_TO_U32( data[0], data[1], data[2], data[3] );
   g.sclk_subseconds = U8_TO_U16( data[4], data[5] );

   /* 
   Assign packet status -- bytes 6 and 7 of the data block.
   */
   g.APID_status     = U8_TO_U16( data[6], data[7] );
   
   /* 
   Assign start level -- bytes 8 and 9, -- and step size -- bytes 10
   and 11.
   */
   g.start_level     = U8_TO_U16( data[8], data[9] );
   g.step_size       = U8_TO_U16( data[10], data[11] );

   /*
   Assign set and level for each of ten levels during one spin. These 
   come in subsequent blocks of 4 bytes. In each block:
      bytes  0..1  -- set;
      byte   2..3  -- level;
   First block starts at byte 12 of the data.
   */  
   for ( i=0; i < GIM_SPIN_LEVELS; i++ )
   {
      /*
      Calculate offset for the first byte of each value in i-th 
      level and assign the structure values.
      */
      setoff = 12 + 4 * i;
      lvloff = 12 + 4 * i + 2;

      g.set[i]    = U8_TO_U16(  data[setoff], data[setoff+1]);
      g.level[i]  = U8_TO_U16(  data[lvloff], data[lvloff+1]);
   }
   
   /*
   Assign CEM_set, CEM_level, AMP_set, MON_5V, MON_12V, SW_status and  
   CMD_status. These come as two byte words starting with byte 32 of
   the data block in the order they defined in the structure.
   */
   g.CEM_set      = U8_TO_U16( data[52], data[53] );
   g.CEM_level    = U8_TO_U16( data[54], data[55] );
   g.AMP_set      = U8_TO_U16( data[56], data[57] );
   g.MON_5V       = U8_TO_U16( data[58], data[59] );
   g.MON_12V      = U8_TO_U16( data[60], data[61] );
   g.SW_status    = U8_TO_U16( data[62], data[63] );
   g.CMD_status   = U8_TO_U16( data[64], data[65] );

   /*
   We finished packaging byte block into the structure and can write
   it to the file.
   */
   if ( write_gim_hk( g, -1 ) == FAIL ) {
      printf(  "ERROR: write_gim_hk() failed\n" );
      exit(-1);
   }
   printf( "      wrote GIM h/k record\n" );

}

/**********************************************************************/
/**********************************************************************/
/**********************************************************************/

/*
   This function packages block of bytes representing 
   complete Concentrator h/k APID and writes it to HDF output file.
*/

int put_CON_hk ( uint8 * data )
{
   /*
   This is the structure that we need to fill out.
   */
   struct Concentrator_Housekeeping g;
   
   /*
   Let get to work. Note, that all offsets for the input data block 
   are hard coded. They correspond to the Concentrator h/k APID layout
   provided by Karl Lauffer on June 19, 2000.
   */

   /* 
   First, assign time  -- first 6 bytes the data block.
   */
   g.sclk_seconds    = U8_TO_U32( data[0], data[1], data[2], data[3] );
   g.sclk_subseconds = U8_TO_U16( data[4], data[5] );

   /* 
   Assign packet status -- bytes 6 and 7 of the data block.
   */
   g.APID_status     = U8_TO_U16( data[6], data[7] );
   
   /*
   Assign the rest of the fields from the structure. The come 
   as two byte words in the order they defined in the structure.
   */
   g.elec_req           = U8_TO_U16( data[8],  data[9]  );
   g.accel_req          = U8_TO_U16( data[10], data[11] );
   g.rej_req            = U8_TO_U16( data[12], data[13] );
   g.elec_level         = U8_TO_U16( data[14], data[15] );
   g.accel_level        = U8_TO_U16( data[16], data[17] );
   g.rej_level          = U8_TO_U16( data[18], data[19] );
   g.elec_set           = U8_TO_U16( data[20], data[21] );
   g.accel_set          = U8_TO_U16( data[22], data[23] );
   g.rej_set            = U8_TO_U16( data[24], data[25] );
   g.MON_5V             = U8_TO_U16( data[26], data[27] );
   g.MON_12V            = U8_TO_U16( data[28], data[29] );
   g.elec_fault_count   = U8_TO_U16( data[30], data[31] );
   g.accel_fault_count  = U8_TO_U16( data[32], data[33] );
   g.rej_fault_count    = U8_TO_U16( data[34], data[35] );
   g.elec_maximum       = U8_TO_U16( data[36], data[37] );
   g.elec_minimum       = U8_TO_U16( data[38], data[39] );
   g.accel_maximum      = U8_TO_U16( data[40], data[41] );
   g.accel_minimum      = U8_TO_U16( data[42], data[43] );
   g.rej_maximum        = U8_TO_U16( data[44], data[45] );
   g.rej_minimum        = U8_TO_U16( data[46], data[47] );
   g.SW_status          = U8_TO_U16( data[48], data[49] );
   g.CMD_status         = U8_TO_U16( data[50], data[51] );

   /*
   We finished packaging byte block into the structure and can write
   it to the file.
   */
   if ( write_con_hk( g, -1 ) == FAIL ) {
      printf(  "ERROR: write_con_hk() failed\n" );
      exit(-1);
   }
   printf( "      wrote Concentrator h/k record\n" );

}

/**********************************************************************/
/**********************************************************************/
/**********************************************************************/

/*
   This function packages block of bytes representing 
   complete MEC/WIND h/k APID and writes it to HDF output file.
*/

int put_MEC_hk ( uint8 * data )
{
   /*
   This is the structure that we need to fill out.
   */
   struct MEC_Housekeeping g;
   
   /*
   Let get to work. Note, that all offsets for the input data block 
   are hard coded. They correspond to the MEC/WIND h/k APID layout
   provided by Karl Lauffer on June 19, 2000 + description of the 
   last field (extra_data) provided by David Gingerich on April 6,
   2001.
   */

   /* 
   First, assign time  -- first 6 bytes the data block.
   */
   g.sclk_seconds    = U8_TO_U32( data[0], data[1], data[2], data[3] );
   g.sclk_subseconds = U8_TO_U16( data[4], data[5] );

   /*
   Assign the rest of the fields of the structure. Since this packet 
   is mapped 1:1 there is no need in any rearranging/re-packing.
   */
   g.WIND_MEC_FLAGS        = U8_TO_U16( data[6], data[7] );
   g.gim_peak_azim         = data[8];
   g.gim_peak_CEM          = data[9];
   g.bde_disqual_cnt       = data[10];
   g.bde_azim_cnt          = data[11];
   g.bde_polar_cnt         = data[12];
   g.regime                = data[13];
   g.WIND_REGIME_FLAGS     = U8_TO_U16( data[14], data[15] );
   C32_TO_F32( data + 16,  g.gim_deadtime );
   C32_TO_F32( data + 20,  g.gim_bg_level );
   C32_TO_F32( data + 24,  g.gem_deadtime );
   C32_TO_F32( data + 28,  g.gem_bg_level );
   C32_TO_F32( data + 32,  g.proton_density_curr );
   C32_TO_F32( data + 36,  g.proton_speed_curr );
   C32_TO_F32( data + 40,  g.proton_temp_curr );
   C32_TO_F32( data + 44,  g.proton_temp_exp_ratio_curr );
   C32_TO_F32( data + 48,  g.alpha_ratio_curr );
   C32_TO_F32( data + 52,  g.proton_density_av );
   C32_TO_F32( data + 56,  g.proton_speed_av );
   C32_TO_F32( data + 60,  g.proton_temp_av );
   C32_TO_F32( data + 64,  g.proton_temp_exp_ratio_av );
   C32_TO_F32( data + 68,  g.alpha_ratio_av );
   C32_TO_F32( data + 72,  g.BDE_av );
   C32_TO_F32( data + 76,  g.TO_CME );
   C32_TO_F32( data + 80,  g.V1_ave );
   C32_TO_F32( data + 84,  g.V2_ave );
   C32_TO_F32( data + 88,  g.N1_ave );
   C32_TO_F32( data + 92,  g.N2_ave );
   C32_TO_F32( data + 96,  g.T1_ave );
   C32_TO_F32( data + 100, g.T2_ave );
   g.TIME_OF_SHOCK         = U8_TO_U32( data[104], data[105], 
                                        data[106], data[107] );
   g.TIME_OF_CME           = U8_TO_U32( data[108], data[109], 
                                        data[110], data[111] );
   C32_TO_F32( data + 112, g.F1D_PROT_PEAK );
   C32_TO_F32( data + 116, g.F1D_MIN );
   C32_TO_F32( data + 120, g.F1D_HE_PEAK );
   g.PROT_PEAK_LEVEL       = U8_TO_U16( data[124], data[125] );
   g.PROT_ALPHA_MIN_LEVEL  = U8_TO_U16( data[126], data[127] );
   g.ALPHA_PEAK_LEVEL      = U8_TO_U16( data[128], data[129] );
   g.HIST_ENTRIES          = U8_TO_U16( data[130], data[131] );
   C32_TO_F32( data + 132, g.HIST_MAX_GAP );
   C32_TO_F32( data + 136, g.COUNT_PEAK );
   g.extra_data            = U8_TO_U32( data[140], data[141], 
                                        data[142], data[143] );

   /*
   We finished packaging byte block into the structure and can write
   it to the file.
   */
   if ( write_mec_hk( g, -1 ) == FAIL ) {
      printf(  "ERROR: write_mec_hk() failed\n" );
      exit(-1);
   }
   printf( "      wrote MEC/WIND h/k record\n" );

}

/**********************************************************************/
/**********************************************************************/
/**********************************************************************/

/*
   This function packages block of bytes representing 
   complete MEC/WIND F1D APID and writes it to HDF output file.
*/

int put_MEC_f1d ( uint8 * data )
{
   /*
   This is the structure that we need to fill out.
   */
   struct MEC_f1d g;
   
   /*
   These are misc counters and offsets.
   */
   int i;
   int offset;
   
   /*
   Let get to work. Note, that all offsets for the input data block 
   are hard coded. They correspond to the MEC/WIND F1D APID layout
   provided by Karl Lauffer on August 14, 2000.
   */

   /* 
   First, assign time  -- first 6 bytes of the data block.
   */
   g.sclk_seconds    = U8_TO_U32( data[0], data[1], data[2], data[3] );
   g.sclk_subseconds = U8_TO_U16( data[4], data[5] );
   
   /* 
   Next two bytes -- bytes 6 & 7 -- are pad.
   */

   /*
   And finish by assigning the array of WIND F1D ion distribution data.
   */
   for ( i=0; i < MEC_F1D_ELEMENTS; i++ )
   {
      /*
      Calculate offset for the first byte of each element 
      and assign the structure values.
      */
      offset = 8 + 4 * i;

      C32_TO_F32( data + offset,  g.f1d_ion_distr[i] );
   }


   /*
   We finished packaging byte block into the structure and can write
   it to the file.
   */
   if ( write_mec_f1d( g, -1 ) == FAIL ) {
      printf(  "ERROR: write_mec_f1d() failed\n" );
      exit(-1);
   }
   printf( "      wrote MEC/WIND F1D record\n" );

}

/**********************************************************************/
/**********************************************************************/
/**********************************************************************/

int parse_command_line( int    argc, 
                        char * argv[], 
                        int  * instrument,
                        int  * apid_ID,
                        char * tlm_file,
                        char * hdf_file )
{
   /*
   Misc. counters
   */
   int clkey;
   
   /*
   Initialize outputs.
   */
   *instrument = 0;
   *apid_ID = 0;
   sprintf( tlm_file, "" );
   sprintf( hdf_file, "" );
   
   /*
   If command line is blank, display usage.
   */
   if ( argc < 3 )
   {
      printf( "\n"                                                );
      printf( "Genesis EDR Generator (%s), %s\n", 
              PROGNAME, PROGVER                                   );
      printf( "\n"                                                );
      printf( "Usage:\n"                                          );
      printf( "\n"                                                );
      printf( "  > %s  %s %s|%s|%s|%s|%s\n",
         PROGNAME, CLKINST, CLKGEM, CLKGIM, CLKCON, CLKMEC, CLKACS );
      printf( "            [%s %s|%s|%s|%s|%s|%s|%s]\n",
         CLKAPID, CLKNORM, CLKHK, CLKMAN, CLKDIAG, CLKRAW, CLKF1D, CLKQUAT );
      printf( "             %s tlm_file\n", CLKTLMF               );
      printf( "             %s hdf_file\n", CLKEDRF               );
      printf( "\n"                                                );
      exit(-1);
   }
   
   /*
   Process command line keys -- instrument goes first. It is required.
   */
   clkey = 1;
            
   if ( ! strcmp( argv[clkey], CLKINST ) )
   {
      if      ( ! strcmp( argv[clkey+1], CLKGEM ) ) *instrument = GEM;
      else if ( ! strcmp( argv[clkey+1], CLKGIM ) ) *instrument = GIM;
      else if ( ! strcmp( argv[clkey+1], CLKCON ) ) *instrument = CON;
      else if ( ! strcmp( argv[clkey+1], CLKMEC ) ) *instrument = MEC;
      else if ( ! strcmp( argv[clkey+1], CLKACS ) ) *instrument = ACS;
      else
      {
         printf(  
                  "ERROR: Cannot recognize instrument ``%s''.\n", 
                  argv[clkey+1]                              );
         exit(-1);
      }
      if ( argc > clkey + 2 )
      {
         clkey = clkey + 2;
      }
      else 
      {
         printf(  "ERROR: Too few command line arguments.\n" );
         exit(-1);
      }
   }
   else
   {
      printf(  "ERROR: Bad or out-of-order option ``%s''.\n", 
               argv[clkey]                                );
      exit(-1);
   }


   /*
   Next we process APID id. It's optional.
   */
   if ( ( strcmp( argv[clkey], CLKAPID ) ) &&
        ( strcmp( argv[clkey], CLKTLMF ) )    )
   {
      printf(  "ERROR: Bad or out-of-order option ``%s''.\n", 
               argv[clkey]                                );
      exit(-1);
   }
         
   if ( ! strcmp( argv[clkey], CLKAPID ) )
   {
      if ( *instrument ) 
      {
         if      ( ! strcmp( argv[clkey+1], CLKHK   ) )
         {
            switch ( *instrument )
            {
               case GEM: *apid_ID = GEM_HK_APID; break;
               case GIM: *apid_ID = GIM_HK_APID; break;
               case CON: *apid_ID = CON_HK_APID; break;
               case MEC: *apid_ID = MEC_HK_APID; break;
               case ACS:
                  printf(  
                     "ERROR: No such APID for this instrument.\n");
                  exit(-1);
            }
         }
         else if ( ! strcmp( argv[clkey+1], CLKNORM ) )
         {
            switch ( *instrument )
            {
               case GEM: *apid_ID = GEM_NORMAL_APID;  break;
               case GIM: *apid_ID = GIM_NORMAL_APID;  break;
               case CON:
               case MEC:
               case ACS:
                  printf(  
                     "ERROR: No such APID for this instrument.\n");
                  exit(-1);
            }
         }
         else if ( ! strcmp( argv[clkey+1], CLKMAN  ) )
         {
            switch ( *instrument )
            {
               case GEM: *apid_ID = GEM_MANCAL_APID;  break;
               case GIM: *apid_ID = GIM_MANCAL_APID;  break;
               case CON:
               case MEC:
               case ACS:
                  printf(  
                     "ERROR: No such APID for this instrument.\n");
                  exit(-1);
            }
         }
         else if ( ! strcmp( argv[clkey+1], CLKRAW  ) )
         {
            switch ( *instrument )
            {
               case GEM: *apid_ID = GEM_RAW_APID;  break;
               case GIM: *apid_ID = GIM_RAW_APID;  break;
               case CON:
               case MEC:
               case ACS:
                  printf(  
                     "ERROR: No such APID for this instrument.\n");
                  exit(-1);
            }
         }
         else if ( ! strcmp( argv[clkey+1], CLKDIAG ) )
         {
            switch ( *instrument )
            {
               case GIM: *apid_ID = GIM_DIAG_APID;  break;
               case GEM:
               case CON:
               case MEC:
               case ACS:
                  printf(  
                     "ERROR: No such APID for this instrument.\n");
                  exit(-1);
            }
         }
         else if ( ! strcmp( argv[clkey+1], CLKF1D ) )
         {
            switch ( *instrument )
            {
               case GIM:
               case GEM:
               case CON:
               case ACS:
                  printf(  
                     "ERROR: No such APID for this instrument.\n");
                  exit(-1);
               case MEC: *apid_ID = MEC_F1D_APID;  break;
            }
         }
         else if ( ! strcmp( argv[clkey+1], CLKQUAT ) )
         {
            switch ( *instrument )
            {
               case GIM:
               case GEM:
               case CON:
               case MEC:
                  printf(  
                     "ERROR: No such APID for this instrument.\n");
                  exit(-1);
               case ACS: *apid_ID = STRK_QUAT_APID;  break;
            }
         }
         else
         {
            printf(  
                     "ERROR: Cannot recognize APID type ``%s''.\n", 
                     argv[clkey+1]                             );
            exit(-1);
         }
         if ( argc > clkey + 2 )
         {
            clkey = clkey + 2;
         }
         else 
         {
            printf(  
                     "ERROR: Too few command line arguments.\n" );
            exit(-1);
         }
      }
      else 
      {
         printf(  
                  "ERROR: Cannot specify APID without instrument.\n" );
         exit(-1);
      }
   }
   
   /*
   Now input file name.
   */
   if ( ! strcmp( argv[clkey], CLKTLMF  ) )
   {
      sprintf( tlm_file, "%s", argv[clkey+1] );
      if ( argc > clkey + 2 )
      {
         clkey = clkey + 2;
      }
      else 
      {
         printf(  "ERROR: Too few command line arguments.\n" );
         exit(-1);
      }
   }
   else 
   {
      printf(  "ERROR: Bad or out-of-order option ``%s''.\n", 
               argv[clkey]                                );
      exit(-1);
   }
         
   /*
   Finally output file name.
   */
   if ( ! strcmp( argv[clkey], CLKEDRF  ) )
   {
      sprintf( hdf_file, "%s", argv[clkey+1] );
   }
   else 
   {
      printf(  "ERROR: Bad or out-of-order option ``%s''.\n", 
               argv[clkey]                                );
      exit(-1);
   }
         
}

/**********************************************************************/
/**********************************************************************/
/**********************************************************************/
