
/*
  
   GNS EDR Structure definitions -- MEC/WIND Housekeeping Data 
  
   BVS/NAIF, April 12, 2001
   
      Added one more unit32 field "extra_data" at the end of the 
      structure to accommodate MEC/WIND APID change (LMA class 
      I/II change ref No. XH0304 from March 2001)
  
   BVS/NAIF, August 16, 2000
   
      Changed COUNT_PEAK type from uint32 to float32 to correspond 
      to the APID structure definition (GenMap.xls spreadsheet by
      Karl Lauffer, LMA.)
  
   BVS/NAIF, June 7, 2000
  
      Changed sclk_subseconds type from uint8 to uint16 to preserve 
      full resolution of the fractional part of SCLK. Updated 
      description to reflect this change. Added WIND_MEC_FLAGS and 
      WIND_REGIME_FLAGS bit breakdown description. Changed COUNT_PEAK
      type from uint16 to uint32 to comply with Karl Lauffers table
      (6/19/00)
  
   BVS/NAIF, May 7, 2000
  
      Changed to correspond to the "better-defined" packet structure
      provided by John Steinberg on April 27, 2000)
  
   BVS/NAIF, March 2, 2000
  
      Reformatted comments; included gnd_edr_const.h.
  
   BVS/NAIF, February 17, 2000
  
*/

#include "hdfi.h"
#include "gns_edr_const.h"

/*
   
   MEC/WIND Housekeeping data structure -- APID # 24
   
      sclk_seconds      -- S/C time at start of data cycle, seconds
      
      sclk_subseconds   -- S/C time at start of data cycle, fractions 
                           (1/65536th) of SCLK second
      
      WIND_MEC_FLAGS    -- MEC status word (MSB order):
      
                              bit 10 = DATA_OK
                              bit 9  = gim_ok
                              bit 8  = gem_ok
                              bit 7  = mec_ok
                              bit 6  = gim_dt_ok
                              bit 5  = peak_find_ok
                              bit 4  = alpha_avail
                              bit 3  = bde_avail
                              bit 2  = bde_detected
                              bit 1  = gem_dt_ok
                              bit 0  = POLAR_EARTH
      
      gim_peak_azim     -- Index on GIM az angle, 0-39
      
      gim_peak_CEM      -- Index on GIM CEM, 0-7
      
      bde_disqual_cnt   -- Number double-peaks in az disqualified 
                           because direction aligned with Earth
                           
      bde_azim_cnt      -- Number double-peaks in az found and Not 
                           disqualified because direction aligned    
                           with Earth
                           
      bde_polar_cnt     -- Number of times clear valleys found in the 
                           polar BDE check (CEM 0 and 6 peaks)
                           
      regime            -- Final result of the regime code - state 
                           that array is commanded to - E, H, or L
                           
      WIND_REGIME_FLAGS -- REGIME status word (MSB order):
      
                              bit 9 = proton_hist_despike
                              bit 8 = alpha_hist_despike
                              bit 7 = conc_to_standby
                              bit 6 = gim_search_req
                              bit 5 = max_gap_exceeded
                              bit 4 = min_count_missed
                              bit 3 = bulk_selected
                              bit 2 = shock_detected
                              bit 1 = alpha weight
                              bit 0 = BDE weight
                              
      gim_deadtime      -- GIM deadtime - the configured parameter, 
                           (UNITS?)
                            
      gim_bg_level      -- GIM background level from on-board 
                           calculation, decimal counts
                           
      gem_deadtime      -- GEM deadtime - the configured parameter,
                           (UNITS?)
                           
      gem_bg_level      -- GEM background level from on-board 
                           calculation, decimal counts
                           
      proton_density_curr  
                        -- proton_density in cm-3
      
      proton_speed_curr -- proton_speed in cm/s 
      
      proton_temp_curr  -- proton_temp in K
      
      proton_temp_exp_ratio_curr 
                        -- proton_temp_exp_ratio, unitless ratio
                                    
      alpha_ratio_curr  -- alpha_ratio, unitless ratio
      
      proton_density_av -- proton_density average over the history,
                           cm-3  
                           
      proton_speed_av   -- proton_speed averaged over the history,
                           cm/s
                            
      proton_temp_av    -- proton_temp_av averaged over the history,
                           units K
                            
      proton_temp_exp_ratio_av   
                        -- proton_temp_exp_ratio_av averaged over the 
                           history, unitless ratio
                                    
      alpha_ratio_av    -- alpha_ratio_av averaged over the history,
                           unitless ratio
                            
      BDE_av            -- BDE average over the history, between 0.0-1.0
      
      TO_CME            -- regime algorithm parameter, between 0.0 and
                           1.0
                           
      V1_ave            -- speed ave over 1st half of history, cm/s
      
      V2_ave            -- speed ave over 2nd half of history, cm/s
      
      N1_ave            -- prot density ave over 1st half of history,
                           cm-3
                           
      N2_ave            -- prot density ave over end half of history,
                           cm-3
                           
      T1_ave            -- temp ave over 1st half of history, K
      
      T2_ave            -- temp ave over 2nd half of history, K
      
      TIME_OF_SHOCK     -- Time of shock, SCLK seconds
      
      TIME_OF_CME       -- time of CME, SCLK seconds
      
      F1D_PROT_PEAK     -- 1D Dist value at proton peak 
      
      F1D_MIN           -- ID Dist value at Min between Proton 
                           and Alpha Peaks
                           
      F1D_HE_PEAK       -- 1D Dist value at the alpha peak
      
      PROT_PEAK_LEVEL   -- energy step (0-39) where peak found
      
      PROT_ALPHA_MIN_LEVEL 
                        -- energy step (0-39) where min found
      
      ALPHA_PEAK_LEVEL  -- energy step (0-39) where alpha peak found
      
      HIST_ENTRIES      -- Number of valid history entries
      
      HIST_MAX_GAP      -- Longest contiguous history gap, (UNITS?)
      
      COUNT_PEAK        -- uncompressed count associated with the peak

      extra_data        -- Extra Wind data and room for expansion
                           (MSB order):

                              bit 31    = aband_data_cycle_bool
                              bit 29-30 = aband_data_cycle_spin(*)
                              bit 0-28  = spare

                           (*) aband_data_cycle_spin contents are 
                              meaningful only when the 
                              aband_data_cycle_bool flag is TRUE.
   
*/

struct MEC_Housekeeping {

   uint32   sclk_seconds;
   uint16   sclk_subseconds;
   uint16   WIND_MEC_FLAGS;
   uint8    gim_peak_azim;
   uint8    gim_peak_CEM;
   uint8    bde_disqual_cnt;
   uint8    bde_azim_cnt;
   uint8    bde_polar_cnt;
   uint8    regime;
   uint16   WIND_REGIME_FLAGS;
   float32  gim_deadtime;
   float32  gim_bg_level;
   float32  gem_deadtime;
   float32  gem_bg_level;
   float32  proton_density_curr;
   float32  proton_speed_curr;
   float32  proton_temp_curr;
   float32  proton_temp_exp_ratio_curr;
   float32  alpha_ratio_curr;
   float32  proton_density_av;
   float32  proton_speed_av;
   float32  proton_temp_av;
   float32  proton_temp_exp_ratio_av;
   float32  alpha_ratio_av;
   float32  BDE_av;
   float32  TO_CME;
   float32  V1_ave;
   float32  V2_ave;
   float32  N1_ave;
   float32  N2_ave;
   float32  T1_ave;
   float32  T2_ave;
   uint32   TIME_OF_SHOCK;
   uint32   TIME_OF_CME;
   float32  F1D_PROT_PEAK;
   float32  F1D_MIN;
   float32  F1D_HE_PEAK;
   uint16   PROT_PEAK_LEVEL;
   uint16   PROT_ALPHA_MIN_LEVEL;
   uint16   ALPHA_PEAK_LEVEL;
   uint16   HIST_ENTRIES;
   float32  HIST_MAX_GAP;
   float32  COUNT_PEAK;
   uint32   extra_data;
                                       
};

